# This file contains functions for plotting

import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import datetime


def visualize_step(env):
    '''Renders the current state of the environment in the terminal'''

    print(f"\n Step: {env.current_step}" +
          f" | {str(env.sim_date.weekday())} {env.sim_date.hour:2d}:{env.sim_date.minute:2d}:{env.sim_date.second:2d} |" +
          f" \tEVs +{env.current_ev_arrived} / -{env.current_ev_departed}" +
          f" | Total: {env.current_evs_parked} / {env.number_of_ports}")

    if env.verbose:
        for cs in env.charging_stations:
            print(f'  - Charging station {cs.id}:')
            print(f'\t Power: {cs.current_power_output:4.1f} kW |' +
                  f' \u2197 {env.charge_prices[cs.id, env.current_step -1 ]:4.2f} €/kWh ' +
                  f' \u2198 {env.discharge_prices[cs.id, env.current_step - 1]:4.2f} €/kWh |' +
                  f' EVs served: {cs.total_evs_served:3d} ' +
                  f' {cs.total_profits:4.2f} €')

            for port in range(cs.n_ports):
                ev = cs.evs_connected[port]
                if ev is not None:
                    print(f'\t\tPort {port}: {ev}')
                else:
                    print(f'\t\tPort {port}:')
        print("")
        for tr in env.transformers:
            print(tr)

        # print current current power setpoint
        print(f' Power setpoints (prev-step): {env.current_power_usage[env.current_step - 1]:.1f} Actual/' +
              f' {env.power_setpoints[env.current_step - 1]:.1f} Setpoint/'
              f' {env.charge_power_potential[env.current_step - 1]:.1f} Potential in kW')


def ev_city_plot(env):
    '''Plots the simulation data

    Plots:
        - The total power and current of each transformer
        - The current of each charging station
        - The energy level of each EV in charging stations
        - The total power of the CPO
    '''
    print("Plotting simulation data at ./results/" + env.sim_name + "/")

    date_range = pd.date_range(start=env.sim_starting_date,
                               end=env.sim_starting_date +
                               (env.simulation_length - 1) *
                               datetime.timedelta(
                                   minutes=env.timescale),
                               freq=f'{env.timescale}min')
    date_range_print = pd.date_range(start=env.sim_starting_date,
                                     end=env.sim_date,
                                     periods=10)
    plt.close('all')
    # close plt ion
    plt.ioff()
    plt.close('all')
    
    # light weight plots when there are too many charging stations
    if not env.lightweight_plots:
        # Plot the energy level of each EV for each charging station
        plt.figure(figsize=(20, 17))        
        # plt.style.use('seaborn-darkgrid')        
        plt.rcParams.update({'font.size': 16})
        plt.rcParams['font.family'] = ['serif']
        
        counter = 1
        dim_x = int(np.ceil(np.sqrt(env.cs)))
        dim_y = int(np.ceil(env.cs/dim_x))
        for cs in env.charging_stations:

            plt.subplot(dim_x, dim_y, counter)
            df = pd.DataFrame([], index=date_range)

            for port in range(cs.n_ports):
                df[port] = env.port_energy_level[port, cs.id, :]

            # Add another row with one datetime step to make the plot look better
            df.loc[df.index[-1] +
                   datetime.timedelta(minutes=env.timescale)] = df.iloc[-1]

            for port in range(cs.n_ports):
                for i, (t_arr, t_dep) in enumerate(env.port_arrival[f'{cs.id}.{port}']):
                    t_dep = t_dep + 1
                    if t_dep > len(df):
                        t_dep = len(df)
                    # x = df.index[t_arr:t_dep]
                    y = df[port].values.T[t_arr:t_dep]
                    # fill y with 0 before and after to match the length of df
                    y = np.concatenate(
                        [np.zeros(t_arr), y, np.zeros(len(df) - t_dep)])

                    plt.step(df.index, y, where='post')
                    plt.fill_between(df.index,
                                     y,
                                     step='post',
                                     alpha=0.7,
                                     label=f'EV {i}, Port {port}')

            plt.title(f'Charging Station {cs.id}', fontsize=24)
            plt.xlabel(f'Time', fontsize=24)
            plt.ylabel('State of Charge (0-1)', fontsize=24)
            plt.xlim([env.sim_starting_date, env.sim_date])
            plt.xticks(ticks=date_range_print,
                       labels=[f'{d.hour:2d}:{d.minute:02d}' for d in date_range_print], rotation=45,
                       fontsize=22)
            # if len(env.port_arrival[f'{cs.id}.{port}']) < 6:
            if dim_x < 3:
                plt.legend()
            plt.grid(True, which='minor', axis='both')
            counter += 1

        plt.tight_layout()
        # Save plt to html
        fig_name = f'results/{env.sim_name}/EV_Energy_Level.png'  # .html
        # plt.show()
        # save in pdf format
        plt.savefig(fig_name, format='png',  # svg
                    dpi=60, bbox_inches='tight')
        
        plt.close('all')
        # Plot the charging and discharging prices
        plt.figure(figsize=(20, 17))

        df = pd.DataFrame([], index=date_range)
        df['charge'] = - env.charge_prices[0, :]
        df['discharge'] = env.discharge_prices[0, :]
        plt.plot(df['charge'], label='Charge prices (€/kW))')
        plt.plot(df['discharge'], label='Discharge prices (€/kW))')
        # plot y = 0 line
        plt.plot([env.sim_starting_date, env.sim_date], [0, 0], 'black')
        plt.legend(fontsize=24)
        plt.grid(True, which='major', axis='both')
        plt.ylabel('Price (€/kW)', fontsize=24)
        plt.xlabel('Time', fontsize=24)
        plt.xlim([env.sim_starting_date, env.sim_date])
        plt.xticks(ticks=date_range_print,
                   labels=[f'{d.hour:2d}:{d.minute:02d}' for d in date_range_print], rotation=45,
                   fontsize=22)
        plt.tight_layout()
        fig_name = f'results/{env.sim_name}/Prices.png'
        plt.savefig(fig_name, format='png',
                    dpi=60, bbox_inches='tight')

        # Plot the total power of each transformer
        plt.figure(figsize=(20, 17))
        counter = 1
        dim_x = int(np.ceil(np.sqrt(env.number_of_transformers)))
        dim_y = int(np.ceil(env.number_of_transformers/dim_x))
        for tr in env.transformers:

            plt.subplot(dim_x, dim_y, counter)
            df = pd.DataFrame([],
                              index=date_range)
            
            colors = plt.cm.gist_earth(np.linspace(0.1, 0.8, len(tr.cs_ids)+1))

            if env.config['inflexible_loads']['include']:
                df['inflexible'] = env.tr_inflexible_loads[tr.id, :] * 1000 / 400
                blue = np.array([0.529, 0.808, 0.922, 1])                
                colors = np.insert(colors, 0, blue, axis=0)
                
            if env.config['solar_power']['include']:
                df['solar'] = env.tr_solar_power[tr.id, :] * 1000 / 400
                gold = np.array([1, 0.843, 0, 1])
                colors = np.insert(colors, 0, gold, axis=0)
                

            for cs in tr.cs_ids:
                df[cs] = env.cs_current[cs, :]

            # create 2 dfs, one for positive power and one for negative
            df_pos = df.copy()
            df_pos[df_pos <= 0] = 0
            df_neg = df.copy()
            df_neg[df_neg > 0] = 0
            

            # Add another row with one datetime step to make the plot look better
            df_pos.loc[df_pos.index[-1] +
                       datetime.timedelta(minutes=env.timescale)] = df_pos.iloc[-1]
            df_neg.loc[df_neg.index[-1] +
                       datetime.timedelta(minutes=env.timescale)] = df_neg.iloc[-1]

            # plot the positive power
            plt.stackplot(df_pos.index, df_pos.values.T,
                          interpolate=True,
                          step='post',
                          alpha=0.7,
                          colors=colors,
                          linestyle='--')

            df['total'] = df.sum(axis=1)
            # print(df)
            max_current = tr.max_current  # * env.timescale / 60
            min_current = tr.min_current  # * env.timescale / 60
            plt.step(df.index,
                     max_current, where='post', color='r', linestyle='--')
            plt.step(df.index, df['total'], 'darkgreen',
                     where='post', linestyle='--')
            plt.plot(df.index,
                     min_current, 'r--')
            plt.stackplot(df_neg.index, df_neg.values.T,
                          interpolate=True,
                          step='post',
                          colors=colors,
                          alpha=0.7,
                          linestyle='--')
            plt.plot([env.sim_starting_date,
                     env.sim_date], [0, 0], 'black')

            # for cs in tr.cs_ids:
            #     plt.step(df.index, df[cs], 'white', where='post', linestyle='--')
            plt.title(f'Transformer {tr.id}')
            plt.xlabel(f'Time')
            plt.ylabel(f'Current (A)')
            plt.xlim([env.sim_starting_date, env.sim_date])
            plt.xticks(ticks=date_range_print,
                       labels=[f'{d.hour:2d}:{d.minute:02d}' for d in date_range_print], rotation=45)
            if len(tr.cs_ids) < 4:
                if env.config['inflexible_loads']['include']:
                    plt.legend(['Inflexible Loads'] +
                               [f'CS {i}' for i in tr.cs_ids] +
                               ['Circuit Breaker Limit (A)', 'Total Current (A)'])
                else:
                    plt.legend([f'CS {i}' for i in tr.cs_ids] +
                               ['Circuit Breaker Limit (A)', 'Total Current (A)'])
            plt.grid(True, which='minor', axis='both')
            counter += 1

        plt.tight_layout()
        # plt.show()
        fig_name = f'results/{env.sim_name}/Transformer_Current.png'
        plt.savefig(fig_name, format='png',
                    dpi=60, bbox_inches='tight')

        # Plot the power of each charging station
        counter = 1
        plt.figure(figsize=(20, 17))
        dim_x = int(np.ceil(np.sqrt(env.cs)))
        dim_y = int(np.ceil(env.cs/dim_x))
        for cs in env.charging_stations:

            plt.subplot(dim_x, dim_y, counter)
            df = pd.DataFrame([], index=date_range)
            df_signal = pd.DataFrame([], index=date_range)

            for port in range(cs.n_ports):
                df[port] = env.port_current[port, cs.id, :]
                df_signal[port] = env.port_current_signal[port, cs.id, :]
                # create 2 dfs, one for positive power and one for negative
            df_pos = df.copy()
            df_pos[df_pos < 0] = 0
            df_neg = df.copy()
            df_neg[df_neg > 0] = 0

            colors = plt.cm.gist_earth(np.linspace(0.1, 0.8, cs.n_ports))

            # Add another row with one datetime step to make the plot look better
            df_pos.loc[df_pos.index[-1] +
                       datetime.timedelta(minutes=env.timescale)] = df_pos.iloc[-1]
            df_neg.loc[df_neg.index[-1] +
                       datetime.timedelta(minutes=env.timescale)] = df_neg.iloc[-1]

            plt.stackplot(df_pos.index, df_pos.values.T,
                          interpolate=True,
                          step='post',
                          alpha=0.7,
                          colors=colors)

            df['total'] = df.sum(axis=1)
            df_signal['total'] = df_signal.sum(axis=1)

            # plot the power limit
            max_charge_current = cs.max_charge_current  # * env.timescale / 60
            max_discharge_current = cs.max_discharge_current  # * env.timescale / 60
            min_charge_current = cs.min_charge_current  # * env.timescale / 60
            min_discharge_current = cs.min_discharge_current  # * env.timescale / 60
            plt.plot([env.sim_starting_date, env.sim_date],
                     [max_charge_current, max_charge_current], 'r--')
            plt.step(df.index, df['total'], 'darkgreen',
                     where='post', linestyle='--')
            plt.step(df_signal.index, df_signal['total'], 'cyan', where='post', alpha=1,
                     linestyle='--')

            plt.plot([env.sim_starting_date, env.sim_date],
                     [min_charge_current, min_charge_current], 'b--')
            plt.plot([env.sim_starting_date, env.sim_date],
                     [max_discharge_current, max_discharge_current], 'r--')
            plt.plot([env.sim_starting_date, env.sim_date],
                     [min_discharge_current, min_discharge_current], 'b--')

            plt.stackplot(df_neg.index, df_neg.values.T,
                          interpolate=True,
                          step='post',
                          colors=colors,
                          alpha=0.7)

            plt.plot([env.sim_starting_date,
                     env.sim_date], [0, 0], 'black')

            # for i in range(cs.n_ports):
            #     plt.step(df.index, df[i], 'grey', where='post', linestyle='--')

            plt.title(f'Charging Station {cs.id}', fontsize=24)
            plt.xlabel(f'Time', fontsize=24)
            plt.ylabel(f'Current (A)', fontsize=24)
            plt.ylim([max_discharge_current*1.1, max_charge_current*1.1])
            plt.xlim([env.sim_starting_date, env.sim_date])
            plt.xticks(ticks=date_range_print,
                       labels=[f'{d.hour:2d}:{d.minute:02d}' for d in date_range_print], rotation=45,
                       fontsize=24)
            # place the legend under each plot

            if dim_x < 3:
                plt.legend([f'Port {i}' for i in range(
                    cs.n_ports)] + ['Max. Current',
                                    'Actual Current',
                                    'Current Signal',
                                    'Min. Current'],
                    fontsize=22,)
            plt.grid(True, which='minor', axis='both')
            counter += 1

        plt.tight_layout()
        # Save plt to html
        fig_name = f'results/{env.sim_name}/CS_Current_signals.png'
        plt.savefig(fig_name, format='png', dpi=60, bbox_inches='tight')

    plt.close('all')
    # Plot the total power for each CS group
    df_total_power = pd.DataFrame([], index=date_range)
    plt.figure(figsize=(10, 8))

    counter = 1
    dim_x = int(np.ceil(np.sqrt(env.number_of_transformers)))
    dim_y = int(np.ceil(env.number_of_transformers/dim_x))
    for tr in env.transformers:

        plt.subplot(dim_x, dim_y, counter)
        df = pd.DataFrame([],
                          index=date_range)

        colors = plt.cm.gist_earth(np.linspace(0.1, 0.8, len(tr.cs_ids)))        
        extra_col = 0

        if env.config['inflexible_loads']['include']:
            df['inflexible'] = env.tr_inflexible_loads[tr.id, :]
            extra_col += 1

        if env.config['solar_power']['include']:
            df['solar'] = env.tr_solar_power[tr.id, :]
            gold = np.array([1, 0.843, 0, 1])
            colors = np.insert(colors, 0, gold, axis=0)
            extra_col += 1

        if env.config['inflexible_loads']['include']:
            # grey = np.array([0.5, 0.5, 0.5, 1])
            light_blue = np.array([0.529, 0.808, 0.922, 1])
            colors = np.insert(colors, 0, light_blue, axis=0)

        for cs in tr.cs_ids:
            df[cs] = env.cs_power[cs, :]

        # create 2 dfs, one for positive power and one for negative
        df_pos = df.copy()
        df_pos[df_pos < 0] = 0
        df_neg = df.copy()
        df_neg[df_neg > 0] = 0

        # Add another row with one datetime step to make the plot look better
        df_pos.loc[df_pos.index[-1] +
                   datetime.timedelta(minutes=env.timescale)] = df_pos.iloc[-1]
        df_neg.loc[df_neg.index[-1] +
                   datetime.timedelta(minutes=env.timescale)] = df_neg.iloc[-1]

        # plot the positive power        
        stacks = plt.stackplot(df_pos.index, df_pos.values.T,
                      interpolate=True,
                      step='post',
                      alpha=0.7,
                      colors=colors,
                    #   hatch=['//'],#, '\\'],#  +[' ']*len(tr.cs_ids),
                    # hatch=tuple(['\\'] +[' ']*len(tr.cs_ids)),
                      linestyle='--')
        
        # for index, stack in enumerate(stacks):
        #     if env.config['inflexible_loads']['include']:
                # stack.set_hatch(['//')

        df['total'] = df.sum(axis=1)
        df_total_power[tr.id] = df['total']
        
        if env.config['demand_response']['include']:
            plt.fill_between(df.index,
                             np.array([tr.max_power.max()] * len(df.index)),
                             tr.max_power,
                             step='post',
                             alpha=0.7,
                             color='r',
                             hatch='xx',
                             linestyle='--',
                             linewidth=2)


        plt.step(df.index, df['total'],
                 'darkgreen',
                 where='post',
                #  linestyle='-',
                 linewidth=3)

        plt.step(df.index,
                 #  tr.max_power
                 [tr.max_power.max()] * len(df.index),
                 where='post',
                 color='r',
                 linestyle='--',
                 linewidth=2)

        plt.stackplot(df_neg.index, df_neg.values.T,
                      interpolate=True,
                      step='post',
                      colors=colors,
                      alpha=0.7,
                    #   hatch='-\ ',
                      linestyle='--')
        plt.plot([env.sim_starting_date, env.sim_date], [0, 0], 'black')

        # for cs in tr.cs_ids:
        #     plt.step(df.index, df[cs], 'white', where='post', linestyle='--')

        plt.title(f'Transformer {tr.id}')
        plt.xlabel(f'Time', fontsize=28)
        plt.ylabel(f'Power (kW)', fontsize=28)
        plt.xlim([env.sim_starting_date, env.sim_date])
        plt.xticks(ticks=date_range_print,
                   labels=[f'{d.hour:2d}:{d.minute:02d}' for d in date_range_print],
                   rotation=45,
                   fontsize=28)
        plt.yticks(fontsize=28)

        legend_list = [f'CS {i+1}' for i in tr.cs_ids] + \
            ['Total Power (kW)'] + \
            ['Power limit (kW)']
            
        if env.config['demand_response']['include']:
            legend_list = legend_list[:-2] + ['Demand Response (kW)'] + legend_list[-2:]

        if len(tr.cs_ids) < 4:

            if env.config['solar_power']['include']:
                legend_list = ['Solar Power'] + legend_list

            if env.config['inflexible_loads']['include']:
                legend_list = ['Inflexible Loads'] + legend_list

            plt.legend(legend_list)
        else:
            if env.config['solar_power']['include']:
                legend_list = ['Solar Power'] + legend_list

            if env.config['inflexible_loads']['include']:
                legend_list = ['Inflexible Loads'] + legend_list

            plt.legend(loc='lower right', fontsize=24, labels=legend_list)

        plt.grid(True, which='minor', axis='both')
        counter += 1

    if len(env.transformers) < 10:
        plt.tight_layout()
        fig_name = f'results/{env.sim_name}/Transformer_Aggregated_Power.png'
        plt.savefig(fig_name, format='png',
                    dpi=60, bbox_inches='tight')
    else:
        # clear plt canvas
        plt.close('all')
        
    # plt.show()
    # Plot the total power of the CPO
    plt.figure(figsize=(20, 17))

    # plt.style.use('seaborn-darkgrid')
    plt.rcParams.update({'font.size': 16})
    plt.rcParams['font.family'] = ['serif']

    # create 2 dfs, one for positive power and one for negative
    df_pos = df_total_power.copy()
    df_pos[df_pos < 0] = 0
    df_neg = df_total_power.copy()
    df_neg[df_neg > 0] = 0

    colors = plt.cm.gist_earth(np.linspace(0.1, 0.8, len(env.transformers)))

    # Add another row with one datetime step to make the plot look better
    df_pos.loc[df_pos.index[-1] +
               datetime.timedelta(minutes=env.timescale)] = df_pos.iloc[-1]
    df_neg.loc[df_neg.index[-1] +
               datetime.timedelta(minutes=env.timescale)] = df_neg.iloc[-1]

    df_total_power['total'] = df_total_power.sum(axis=1)

    plt.step(df_total_power.index, df_total_power['total'], 'darkgreen',
             where='post', linestyle='--')

    plt.step(df_total_power.index, env.power_setpoints, 'r--', where='post',)

    if env.load_from_replay_path is not None:
        plt.step(df_total_power.index, env.replay.ev_load_potential,
                 'b--', where='post', alpha=0.4,)
    # else:
    #     plt.step(df_total_power.index, env.current_power_usage,
    #              'b--', where='post', alpha=0.4,)

    # plot the positive power
    plt.stackplot(df_pos.index, df_pos.values.T,
                  interpolate=True,
                  step='post',
                  alpha=0.7,
                  colors=colors,
                  linestyle='--')

    plt.stackplot(df_neg.index, df_neg.values.T,
                  interpolate=True,
                  step='post',
                  colors=colors,
                  alpha=0.7,
                  linestyle='--')

    plt.plot([env.sim_starting_date, env.sim_date], [0, 0], 'black')

    # for cs in tr.cs_ids:
    #     plt.step(df.index, df[cs], 'white', where='post', linestyle='--')
    # plt.title(f'Aggreagated Power vs Power Setpoint', fontsize=44)
    plt.xlabel(f'Time', fontsize=38)
    plt.ylabel(f'Power (kW)', fontsize=38)
    plt.xlim([env.sim_starting_date, env.sim_date])

    date_range_print = pd.date_range(start=env.sim_starting_date,
                                     end=env.sim_date,
                                     periods=7)
    plt.xticks(ticks=date_range_print,
               labels=[f'{d.hour:2d}:{d.minute:02d}' for d in date_range_print], rotation=45, fontsize=28)
    # plt.xticks(ticks=date_range_print,
    #            labels=[f'{d.strftime("%A")}' for d in date_range_print], rotation=45, fontsize=28)

    # set ytick font size
    plt.yticks(fontsize=28)
    if len(env.transformers) <= 10:
        if env.load_from_replay_path is not None:
            plt.legend(['Total Power'] +
                       [f'Power Setpoint'] +
                       ['EV Unsteered Load Potential (kW)']
                       + [f'Tr {i}' for i in range(len(env.transformers))])
        else:
            plt.legend(['Total Power'] +
                       [f'Power Setpoint']
                       + [f'Tr {i}' for i in range(len(env.transformers))])
    else:
        # plt.legend(['Total Power (kW)']+[f'Power Setpoint (kW)']+['EV Unsteered Load Potential (kW)'])
        plt.legend(['Total Power (kW)'], fontsize=28)
    plt.grid(True, which='minor', axis='both')

    plt.tight_layout()
    # plt.show()
    fig_name = f'results/{env.sim_name}/Total_Aggregated_Power.png'
    plt.savefig(fig_name, format='png',
                dpi=60, bbox_inches='tight')

    # plot prices
    # plt.figure(figsize=(20, 17))
    # plt.plot(env.charge_prices[0,:], label='Charge prices (€/kW))')
    # plt.plot(env.discharge_prices[0,:], label='Discharge prices (€/kW))')
    # plt.legend()
    # plt.grid(True, which='minor', axis='both')
    # plt.tight_layout()
    # fig_name = f'results/{env.sim_name}/Prices.png'
    # plt.savefig(fig_name, format='png',
    #             dpi=60, bbox_inches='tight')

    plt.close('all')



# this class contains heurisyic algorithms for the power setpoint tracking problem
import math
import numpy as np
from typing import List


class RoundRobin():
    '''
    This is a class that contains the Round Robin heuristic algorithm for the power setpoint tracking problem.
    It does not consider multiple transfomer constraints. 
    And it assumes all chargers have the same number of ports
    '''
    algo_name = "Round Robin"

    def __init__(self, env, verbose=False,  **kwargs):

        self.verbose = verbose
        self.env = env
        # find average charging power of the simulation
        self.average_power = 0
        for cs in env.charging_stations:
            self.average_power += cs.max_charge_current * \
                cs.voltage * math.sqrt(cs.phases) / cs.n_ports
        self.average_power /= len(env.charging_stations)

        self.number_of_ports_per_cs = env.number_of_ports_per_cs
        # list with the ids of EVs that were already served in this round
        self.ev_buffer = []

    def get_env(self):
        return self.env

    def update_ev_buffer(self, env) -> None:
        '''
        This function updates the EV buffer list with the EVs that are currently parked by adding or removing them.
        '''
        counter = 0
        # iterate over all ports
        for cs in env.charging_stations:
            for port in range(cs.n_ports):
                if cs.evs_connected[port] is not None:
                    if cs.evs_connected[port].get_soc() < 1:

                        if counter not in self.ev_buffer:
                            self.ev_buffer.insert(0, counter)
                    else:
                        if counter in self.ev_buffer:
                            self.ev_buffer.remove(counter)
                else:
                    if counter in self.ev_buffer:
                        self.ev_buffer.remove(counter)
                counter += 1

    def get_action(self, env) -> np.ndarray:

        # this function returns the action list based on the round robin algorithm

        total_power = env.power_setpoints[env.current_step] * 1000  # in W

        number_of_EVs_to_charge = total_power / self.average_power

        if self.verbose:
            print("-------------------Round Robin-------------------")
            print(
                f'Number of EVs to charge: {number_of_EVs_to_charge:.2f},\
                total power: {total_power:.2f}, average power: {self.average_power:.2f}')

        # get currently parked EVs
        self.update_ev_buffer(env)
        max_number_of_EVs_to_charge = len(self.ev_buffer)

        if self.verbose:
            print(f'EV buffer: {self.ev_buffer}')

        # get the EVs to charge in this round
        evs_to_charge = self.ev_buffer[:min(
            int(np.ceil(number_of_EVs_to_charge)), max_number_of_EVs_to_charge)]
        self.ev_buffer = self.ev_buffer[min(
            int(np.ceil(number_of_EVs_to_charge)), max_number_of_EVs_to_charge):]

        self.ev_buffer.extend(evs_to_charge)

        # create action list
        action_list = np.zeros(env.number_of_ports)

        # set the action for the EVs to charge
        for i, ev in enumerate(evs_to_charge):
            action_list[ev] = 1 / env.number_of_ports_per_cs
            if i == len(evs_to_charge) - 1:
                action_list[ev] = (number_of_EVs_to_charge - i)

        if self.verbose:
            print(f'Evs to charge: {evs_to_charge}')

        return action_list


class ChargeAsLateAsPossible():
    '''
    This is a class that contains the Charge As Late As Possible heuristic algorithm.
    '''
    algo_name = "Charge As Late As Possible"

    def __init__(self, verbose=False, **kwargs):

        self.verbose = verbose

    def update_ev_buffer(self, env) -> List[int]:
        '''
        This function updates the EV buffer list with the EVs that are currently parked by adding or removing them.
        '''
        ev_buffer = []
        counter = 0
        # iterate over all ports
        for cs in env.charging_stations:
            for port in range(cs.n_ports):
                if cs.evs_connected[port] is not None:

                    cs_max_power = cs.max_charge_current * \
                        cs.voltage * math.sqrt(cs.phases) / 1000
                    # find minimum steps required to charge the EV
                    min_steps = math.ceil((1 - cs.evs_connected[port].get_soc())
                                          / (cs_max_power * env.timescale/60 / cs.evs_connected[port].battery_capacity))

                    start_of_charging_step = cs.evs_connected[port].time_of_departure - min_steps
                    if cs.evs_connected[port].get_soc() < 1 and start_of_charging_step <= env.current_step:
                        ev_buffer.append(counter)

                counter += 1

        return ev_buffer

    def get_action(self, env) -> np.ndarray:

        # this function returns the action list based on the round robin algorithm

        ev_buffer = self.update_ev_buffer(env)
        # create action list
        action_list = np.zeros(env.number_of_ports)

        # set the action for the EVs to charge
        for i, ev in enumerate(ev_buffer):
            action_list[ev] = 1

        return action_list


class ChargeAsFastAsPossible():
    '''
    This class contains the Charge As Fast As Possible heuristic algorithm.
    '''
    algo_name = "Charge As Fast As Possible"
    
    def __init__(self, verbose=False, **kwargs):
        self.verbose = verbose

    def get_action(self, env) -> np.ndarray:
        '''
        This function returns the action list based on the charge as fast as possible algorithm.
        '''
        action_list = np.ones(env.number_of_ports)
        return action_list


class ChargeAsFastAsPossibleToDesiredCapacity():
    '''
    This class contains the Charge As Fast As Possible heuristic algorithm.
    '''
    algo_name = "Charge As Fast As Possible To Desired Capacity"

    def __init__(self, verbose=False, **kwargs):
        self.verbose = verbose

    def get_action(self, env) -> np.ndarray:
        '''
        This function returns the action list based on the charge as fast as possible algorithm.
        It charges the EVs up to their desired capacity.
        Desired capacity can be different than the maximum capacity of the battery.
        '''
        action_list = np.zeros(env.number_of_ports)

        counter = 0
        for i, cs in enumerate(env.charging_stations):
            for port in range(cs.n_ports):
                if cs.evs_connected[port] is not None:
                    max_power_to_charge_cs = cs.get_max_power()
                    max_power_to_charge = min(
                        max_power_to_charge_cs, cs.evs_connected[port].max_ac_charge_power)

                    max_energy_to_charge = max_power_to_charge * env.timescale/60
                    if cs.evs_connected[port].current_capacity + max_energy_to_charge < cs.evs_connected[port].desired_capacity:
                        action_list[counter] = 1
                    else:
                        action_list[counter] = ((cs.evs_connected[port].desired_capacity -
                                                cs.evs_connected[port].current_capacity) *
                                                60 / env.timescale) / max_power_to_charge_cs
                        if action_list[counter] < 0:
                            action_list[counter] = 0

                counter += 1

        return action_list



import argparse

def arg_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("--env", default="ev-city-v0",
                        help="the environment on which the agent should be trained ")
    parser.add_argument("--name",type=str, help="name of the experiment")
    parser.add_argument("--render_train", default=False, type=bool,
                        help="Render the training steps (default: False)")
    parser.add_argument("--render_eval", default=True, type=bool,
                        help="Render the evaluation steps (default: False)")
    parser.add_argument("--load_model", default=False, type=bool,
                        help="Load a pretrained model (default: False)")
    parser.add_argument("--save_dir", default="./saved_models/",
                        help="Dir. path to save and load a model (default: ./saved_models/)")
    parser.add_argument("--seed", default=42, type=int,
                        help="Random seed (default: 0)")
    parser.add_argument("--timesteps", default=10*1e6, type=int,
                        help="Num. of total timesteps of training (default: 1e6)")
    parser.add_argument("--batch_size", default=512, type=int,#128
                        help="Batch size (default: 64; OpenAI: 128)")
    parser.add_argument("--replay_size", default=1e5, type=int,
                        help="Size of the replay buffer (default: 1e6; OpenAI: 1e5)")
    parser.add_argument("--gamma", default=0.99,
                        help="Discount factor (default: 0.99)")
    parser.add_argument("--tau", default=0.001,
                        help="Update factor for the soft update of the target networks (default: 0.001)")
    parser.add_argument("--noise_stddev", default=0.3, type=int,
                        help="Standard deviation of the OU-Noise (default: 0.2)")
    parser.add_argument("--hidden_size", nargs=2, default=[256, 256], type=tuple,
                        help="Num. of units of the hidden layers (default: [400, 300]; OpenAI: [64, 64])")    
    parser.add_argument("--wandb", default=True, type=bool,
                        help="Enable logging to wandb (default: True)")

    # Environment specific arguments
    parser.add_argument("--config_file", default="ev2gym/example_config_files/PublicPST.yaml",
    # parser.add_argument("--config_file", default="ev2gym/example_config_files/V2G_MPC.yaml",
    # parser.add_argument("--config_file", default="ev2gym/example_config_files/V2GProfitPlusLoads.yaml",
                        help="Path to the config file (default: config_files/config.yaml)")
    parser.add_argument("--n_test_cycles", default=50, type=int,
                        help="Num. of episodes in the evaluation phases (default: 10; OpenAI: 20)")

    # Generate trajectories specific arguments
    parser.add_argument("--n_trajectories", default=200_000, type=int,
                        help="Num. of trajectories to generate (default: 10)")
    
    #
    parser.add_argument("--dataset", default="RR", type=str)
    parser.add_argument("--save_opt_trajectories", default=True, type=bool,
                        help="Save Optimal trajectories (default: False)")
    

    return parser.parse_args()



'''
This file contains the EVCity class, which is used to represent the environment of the city.
The environment is a gym environment and can be also used with the OpenAI gym standards and baselines.
The environment an also be used for standalone simulations without the gym environment.

===================================
Author: Stavros Orfanoudakis 2023
===================================
'''

import gymnasium as gym
from gymnasium import spaces
import numpy as np
import datetime
import pickle
import os
import random
from copy import deepcopy
import yaml
import json

# from .grid import Grid
from ev2gym.models.replay import EvCityReplay
from ev2gym.visuals.plots import ev_city_plot, visualize_step
from ev2gym.utilities.utils import get_statistics, print_statistics, calculate_charge_power_potential
from ev2gym.utilities.loaders import load_ev_spawn_scenarios, load_power_setpoints, load_transformers, load_ev_charger_profiles, load_ev_profiles, load_electricity_prices
from ev2gym.visuals.render import Renderer

from ev2gym.rl_agent.reward import SquaredTrackingErrorReward
from ev2gym.rl_agent.state import PublicPST


class EV2Gym(gym.Env):

    def __init__(self,
                 config_file=None,
                 load_from_replay_path=None,  # path of replay file to load
                 replay_save_path='./replay/',  # where to save the replay file
                 generate_rnd_game=True,  # generate a random game without terminating conditions
                 seed=None,
                 save_replay=False,
                 save_plots=False,
                 state_function=PublicPST,
                 reward_function=SquaredTrackingErrorReward,
                 eval_mode="Normal",  # eval mode can be "Normal", "Unstirred" or "Optimal" in order to save the correct statistics in the replay file
                 lightweight_plots=False,
                 # whether to empty the ports at the end of the simulation or not
                 empty_ports_at_end_of_simulation=True,
                 extra_sim_name=None,
                 verbose=False,
                 render_mode=None,
                 ):

        super(EV2Gym, self).__init__()

        if verbose:
            print(f'Initializing EVs-Simulator environment...')

        # read yaml config file
        assert config_file is not None, "Please provide a config file!!!"
        self.config = yaml.load(open(config_file, 'r'), Loader=yaml.FullLoader)
        
        self.generate_rnd_game = generate_rnd_game
        self.load_from_replay_path = load_from_replay_path
        self.empty_ports_at_end_of_simulation = empty_ports_at_end_of_simulation
        self.save_replay = save_replay
        self.save_plots = save_plots
        self.lightweight_plots = lightweight_plots
        self.eval_mode = eval_mode
        self.verbose = verbose  # Whether to print the simulation progress or not
        # Whether to render the simulation in real-time or not
        self.render_mode = render_mode

        self.simulation_length = self.config['simulation_length']

        self.replay_path = replay_save_path

        cs = self.config['number_of_charging_stations']

        self.reward_function = reward_function
        self.state_function = state_function

        if seed is None:
            self.seed = np.random.randint(0, 1000000)
            # print(f"Random seed: {self.seed}")
        else:
            self.seed = seed
        # set random seed
        np.random.seed(self.seed)
        random.seed(self.seed)
        
        self.tr_seed = self.config['tr_seed']        
        if self.tr_seed == -1:
            self.tr_seed = self.seed            
        self.tr_rng = np.random.default_rng(seed=self.tr_seed)
                

        if load_from_replay_path is not None:
            with open(load_from_replay_path, 'rb') as file:
                self.replay = pickle.load(file)

            sim_name = self.replay.replay_path.split(
                'replay_')[-1].split('.')[0]
            self.sim_name = sim_name + '_replay'
            self.sim_date = self.replay.sim_date
            self.timescale = self.replay.timescale
            self.cs = self.replay.n_cs
            self.number_of_transformers = self.replay.n_transformers
            self.number_of_ports_per_cs = self.replay.max_n_ports
            self.scenario = self.replay.scenario
            self.heterogeneous_specs = self.replay.heterogeneous_specs

        else:
            assert cs is not None, "Please provide the number of charging stations"
            self.cs = cs  # Number of charging stations
            # Threshold for the user satisfaction score

            self.number_of_ports_per_cs = self.config['number_of_ports_per_cs']
            self.number_of_transformers = self.config['number_of_transformers']
            self.timescale = self.config['timescale']
            self.simulation_length = int(self.config['simulation_length'])
            # Simulation time

            self.sim_date = datetime.datetime(self.config['year'],
                                              self.config['month'],
                                              self.config['day'],
                                              self.config['hour'],
                                              self.config['minute'])
            self.replay = None
            self.sim_name = f'sim_' + \
                f'{datetime.datetime.now().strftime("%Y_%m_%d_%f")}'

            self.scenario = self.config['scenario']
            self.heterogeneous_specs = self.config['heterogeneous_ev_specs']

        # Whether to simulate the grid or not (Future feature...)
        self.simulate_grid = False

        if self.cs > 100:
            self.lightweight_plots = True
        self.sim_starting_date = self.sim_date

        # Read the config.charging_network_topology json file and read the topology
        try:
            with open(self.config['charging_network_topology']) as json_file:
                self.charging_network_topology = json.load(json_file)

        except FileNotFoundError:
            if not self.config['charging_network_topology'] == 'None':
                print(
                    f'Did not find file {self.config["charging_network_topology"]}')
            self.charging_network_topology = None

        self.sim_name = extra_sim_name + \
            self.sim_name if extra_sim_name is not None else self.sim_name

        # Simulate grid
        if self.simulate_grid:
            pass
            # self.grid = Grid(charging_stations=self.cs, case=case)
            # self.cs_buses = self.grid.get_charging_stations_buses()
            # self.cs_transformers = self.grid.get_bus_transformers()
        else:
            # self.cs_buses = [None] * self.cs
            if self.charging_network_topology is None:
                self.cs_transformers = [
                    *np.arange(self.number_of_transformers)] * (self.cs // self.number_of_transformers)
                self.cs_transformers += random.sample(
                    [*np.arange(self.number_of_transformers)], self.cs % self.number_of_transformers)
                random.shuffle(self.cs_transformers)

        # Instatiate Transformers
        self.transformers = load_transformers(self)
        for tr in self.transformers:
            tr.reset(step=0)

        # Instatiate Charging Stations
        self.charging_stations = load_ev_charger_profiles(self)
        for cs in self.charging_stations:
            cs.reset()

        # Calculate the total number of ports in the simulation
        self.number_of_ports = np.array(
            [cs.n_ports for cs in self.charging_stations]).sum()

        # Load EV spawn scenarios
        if self.load_from_replay_path is None:
            load_ev_spawn_scenarios(self)

        # Spawn EVs
        self.EVs_profiles = load_ev_profiles(self)
        self.EVs = []

        # Load Electricity prices for every charging station
        self.charge_prices, self.discharge_prices = load_electricity_prices(
            self)

        # Load power setpoint of simulation
        self.power_setpoints = load_power_setpoints(self)
        self.current_power_usage = np.zeros(self.simulation_length)
        self.charge_power_potential = np.zeros(self.simulation_length)

        self.init_statistic_variables()

        # Variable showing whether the simulation is done or not
        self.done = False

        # Make folders for results
        if self.save_replay:
            os.makedirs(self.replay_path, exist_ok=True)

        if self.render_mode:
            # Initialize the rendering of the simulation
            self.renderer = Renderer(self)

        if self.save_plots:
            os.makedirs("./results", exist_ok=True)
            print(f"Creating directory: ./results/{self.sim_name}")
            os.makedirs(f"./results/{self.sim_name}", exist_ok=True)

        # Action space: is a vector of size "Sum of all ports of all charging stations"
        high = np.ones([self.number_of_ports])
        if self.config['v2g_enabled']:
            lows = -1 * np.ones([self.number_of_ports])
        else:
            lows = np.zeros([self.number_of_ports])
        self.action_space = spaces.Box(low=lows, high=high, dtype=np.float64)

        # Observation space: is a matrix of size ("Sum of all ports of all charging stations",n_features)
        obs_dim = len(self._get_observation())

        high = np.inf*np.ones([obs_dim])
        self.observation_space = spaces.Box(
            low=-high, high=high, dtype=np.float64)

        # Observation mask: is a vector of size ("Sum of all ports of all charging stations") showing in which ports an EV is connected
        self.observation_mask = np.zeros(self.number_of_ports)

    def reset(self, seed=None, options=None, **kwargs):
        '''Resets the environment to its initial state'''

        if seed is None:
            self.seed = np.random.randint(0, 1000000)
        else:
            self.seed = seed

        # set random seed
        np.random.seed(self.seed)
        random.seed(self.seed)
        
        if self.tr_seed == -1:
            self.tr_seed = self.seed            
        self.tr_rng = np.random.default_rng(seed=self.tr_seed)

        self.current_step = 0
        # Reset all charging stations
        for cs in self.charging_stations:
            cs.reset()
            
        for tr in self.transformers:
            tr.reset(step=self.current_step)

        if self.load_from_replay_path is not None or not self.config['random_day']:
            self.sim_date = self.sim_starting_date
        else:
            # select random date in range

            self.sim_date = datetime.datetime(2022,
                                              1,
                                              1,
                                              self.config['hour'],
                                              self.config['minute'],
                                              ) + datetime.timedelta(days=random.randint(0, int(1.5*365)))

            if self.scenario == 'workplace':
                # dont simulate weekends
                while self.sim_date.weekday() > 4:
                    self.sim_date += datetime.timedelta(days=1)

            if self.config['simulation_days'] == "weekdays":
                # dont simulate weekends
                while self.sim_date.weekday() > 4:
                    self.sim_date += datetime.timedelta(days=1)
            elif self.config['simulation_days'] == "weekends" and self.scenario != 'workplace':
                # simulate only weekends
                while self.sim_date.weekday() < 5:
                    self.sim_date += datetime.timedelta(days=1)

        self.sim_starting_date = self.sim_date
        self.EVs_profiles = load_ev_profiles(self)
        self.power_setpoints = load_power_setpoints(self)
        self.EVs = []

        # print(f'Simulation starting date: {self.sim_date}')

        # self.sim_name = f'ev_city_{self.simulation_length}_' + \
        # f'{datetime.datetime.now().strftime("%Y-%m-%d_%H-%M")}'

        self.init_statistic_variables()

        return self._get_observation(), {}

    def init_statistic_variables(self):
        '''
        Initializes the variables used for keeping simulation statistics
        '''
        self.current_step = 0
        self.total_evs_spawned = 0
        self.total_reward = 0

        self.current_ev_departed = 0
        self.current_ev_arrived = 0
        self.current_evs_parked = 0

        self.previous_power_usage = self.current_power_usage
        self.current_power_usage = np.zeros(self.simulation_length)

        # self.transformer_amps = np.zeros([self.number_of_transformers,
        #                                   self.simulation_length])

        self.cs_power = np.zeros([self.cs, self.simulation_length])
        self.cs_current = np.zeros([self.cs, self.simulation_length])

        self.tr_overload = np.zeros(
            [self.number_of_transformers, self.simulation_length])

        self.tr_inflexible_loads = np.zeros(
            [self.number_of_transformers, self.simulation_length])

        self.tr_solar_power = np.zeros(
            [self.number_of_transformers, self.simulation_length])

        # self.port_power = np.zeros([self.number_of_ports,
        #                             self.cs,
        #                             self.simulation_length],
        #                            dtype=np.float16)
        if not self.lightweight_plots:
            self.port_current = np.zeros([self.number_of_ports,
                                          self.cs,
                                          self.simulation_length],
                                         dtype=np.float16,
                                         )
            self.port_current_signal = np.zeros([self.number_of_ports,
                                                 self.cs,
                                                 self.simulation_length],
                                                dtype=np.float16,
                                                )

            self.port_energy_level = np.zeros([self.number_of_ports,
                                               self.cs,
                                               self.simulation_length],
                                              dtype=np.float16)
            # self.port_charging_cycles = np.zeros([self.number_of_ports,
            #                                       self.cs,
            #                                       self.simulation_length],
            #                                      dtype=np.float16)
            self.port_arrival = dict({f'{j}.{i}': []
                                      for i in range(self.number_of_ports)
                                      for j in range(self.cs)})

        self.done = False

    def step(self, actions, visualize=False):
        ''''
        Takes an action as input and returns the next state, reward, and whether the episode is done
        Inputs:
            - actions: is a vector of size "Sum of all ports of all charging stations taking values in [-1,1]"
        Returns:
            - observation: is a matrix with the complete observation space
            - reward: is a scalar value representing the reward of the current step
            - done: is a boolean value indicating whether the episode is done or not
        '''
        assert not self.done, "Episode is done, please reset the environment"

        if self.verbose:
            print("-"*80)

        total_costs = 0
        total_invalid_action_punishment = 0
        user_satisfaction_list = []
        departing_evs = []

        self.current_ev_departed = 0
        self.current_ev_arrived = 0

        port_counter = 0

        # Reset current power of all transformers
        for tr in self.transformers:
            tr.reset(step=self.current_step)

        # Call step for each charging station and spawn EVs where necessary
        for i, cs in enumerate(self.charging_stations):
            n_ports = cs.n_ports
            costs, user_satisfaction, invalid_action_punishment, ev = cs.step(
                actions[port_counter:port_counter + n_ports],
                self.charge_prices[cs.id, self.current_step],
                self.discharge_prices[cs.id, self.current_step])

            departing_evs += ev

            for u in user_satisfaction:
                user_satisfaction_list.append(u)

            self.current_power_usage[self.current_step] += cs.current_power_output

            # Update transformer variables for this timestep
            self.transformers[cs.connected_transformer].step(
                cs.current_total_amps, cs.current_power_output)

            total_costs += costs
            total_invalid_action_punishment += invalid_action_punishment
            self.current_ev_departed += len(user_satisfaction)

            port_counter += n_ports

        # Spawn EVs
        counter = self.total_evs_spawned
        for i, ev in enumerate(self.EVs_profiles[counter:]):
            if ev.time_of_arrival == self.current_step + 1:
                ev = deepcopy(ev)
                ev.reset()
                ev.simulation_length = self.simulation_length
                index = self.charging_stations[ev.location].spawn_ev(ev)

                if not self.lightweight_plots:
                    self.port_arrival[f'{ev.location}.{index}'].append(
                        (self.current_step+1, ev.time_of_departure+1))

                self.total_evs_spawned += 1
                self.current_ev_arrived += 1
                self.EVs.append(ev)

            elif ev.time_of_arrival > self.current_step + 1:
                break

        self._update_power_statistics(departing_evs)

        self.current_step += 1
        self._step_date()

        if self.current_step < self.simulation_length:
            self.charge_power_potential[self.current_step] = calculate_charge_power_potential(
                self)

        self.current_evs_parked += self.current_ev_arrived - self.current_ev_departed

        # Call step for the grid
        if self.simulate_grid:
            # TODO: transform actions -> grid_actions
            raise NotImplementedError
            grid_report = self.grid.step(actions=actions)
            reward = self._calculate_reward(grid_report)
        else:
            reward = self._calculate_reward(total_costs,
                                            user_satisfaction_list,
                                            total_invalid_action_punishment)

        if visualize:
            visualize_step(self)

        self.render()

        return self._check_termination(user_satisfaction_list, reward)

    def _check_termination(self, user_satisfaction_list, reward):
        '''Checks if the episode is done or any constraint is violated'''
        truncated = False
        # Check if the episode is done or any constraint is violated
        if self.current_step >= self.simulation_length or \
            (any(tr.is_overloaded() > 0 for tr in self.transformers)
             and not self.generate_rnd_game):
            """Terminate if:
                - The simulation length is reached
                - Any user satisfaction score is below the threshold
                - Any charging station is overloaded
                Dont terminate when overloading if :
                - generate_rnd_game is True
                Carefull: if generate_rnd_game is True,
                the simulation might end up in infeasible problem
                """
            if self.verbose:
                print_statistics(self)

                if any(tr.is_overloaded() for tr in self.transformers):
                    print(
                        f"Transformer overloaded, {self.current_step} timesteps\n")
                else:
                    print(
                        f"Episode finished after {self.current_step} timesteps\n")

            if self.save_replay:
                self._save_sim_replay()

            if self.save_plots:
                #save the env as a pickle file
                with open(f"./results/{self.sim_name}/env.pkl", 'wb') as f:
                    self.renderer = None
                    pickle.dump(self, f)
                ev_city_plot(self)

            self.done = True
            return self._get_observation(), reward, True, truncated, get_statistics(self)
        else:
            return self._get_observation(), reward, False, truncated, {'None': None}

    def render(self):
        '''Renders the simulation'''
        if self.render_mode:
            self.renderer.render()

    def _save_sim_replay(self):
        '''Saves the simulation data in a pickle file'''
        replay = EvCityReplay(self)
        print(f"Saving replay file at {replay.replay_path}")
        with open(replay.replay_path, 'wb') as f:
            pickle.dump(replay, f)

        return replay.replay_path
    
    def set_save_plots(self, save_plots):
        if save_plots:
            os.makedirs("./results", exist_ok=True)
            print(f"Creating directory: ./results/{self.sim_name}")
            os.makedirs(f"./results/{self.sim_name}", exist_ok=True)
            
        self.save_plots = save_plots

    def _update_power_statistics(self, departing_evs):
        '''Updates the power statistics of the simulation'''

        # if not self.lightweight_plots:
        for tr in self.transformers:
            # self.transformer_amps[tr.id, self.current_step] = tr.current_amps
            self.tr_overload[tr.id,
                             self.current_step] = tr.get_how_overloaded()
            self.tr_inflexible_loads[tr.id,
                                     self.current_step] = tr.inflexible_load[self.current_step]
            self.tr_solar_power[tr.id,
                                self.current_step] = tr.solar_power[self.current_step]

        for cs in self.charging_stations:
            self.cs_power[cs.id, self.current_step] = cs.current_power_output
            self.cs_current[cs.id, self.current_step] = cs.current_total_amps

            for port in range(cs.n_ports):
                if not self.lightweight_plots:
                    self.port_current_signal[port, cs.id,
                                             self.current_step] = cs.current_signal[port]
                ev = cs.evs_connected[port]
                if ev is not None and not self.lightweight_plots:
                    # self.port_power[port, cs.id,
                    #                 self.current_step] = ev.current_energy
                    self.port_current[port, cs.id,
                                      self.current_step] = ev.actual_current

                    self.port_energy_level[port, cs.id,
                                           self.current_step] = ev.current_capacity/ev.battery_capacity

            for ev in departing_evs:
                if not self.lightweight_plots:
                    self.port_energy_level[ev.id, ev.location, self.current_step] = \
                        ev.current_capacity/ev.battery_capacity
                    self.port_current[ev.id, ev.location,
                                      self.current_step] = ev.actual_current

    def _step_date(self):
        '''Steps the simulation date by one timestep'''
        self.sim_date = self.sim_date + \
            datetime.timedelta(minutes=self.timescale)

    def _get_observation(self):

        return self.state_function(self)

    def set_reward_function(self, reward_function):
        '''
        This function sets the reward function of the environment
        '''
        self.reward_function = reward_function

    def _calculate_reward(self, total_costs, user_satisfaction_list, invalid_action_punishment):
        '''Calculates the reward for the current step'''

        reward = self.reward_function(self, total_costs, user_satisfaction_list, invalid_action_punishment)
        self.total_reward += reward
        
        return reward



import pandas as pd
import matplotlib.pyplot as plt

#read csv file Netherlands_day-ahead-2015-2023

# data = pd.read_csv(r'.\data\Netherlands_day-ahead-2015-2023.csv', sep=',', header=0)
# drop_columns = ['Country','Datetime (UTC)']
# data.drop(drop_columns, inplace=True, axis=1)
# data['year'] = pd.DatetimeIndex(data['Datetime (Local)']).year
# data['month'] = pd.DatetimeIndex(data['Datetime (Local)']).month
# data['day'] = pd.DatetimeIndex(data['Datetime (Local)']).day
# data['hour'] = pd.DatetimeIndex(data['Datetime (Local)']).hour
# print(data)


#read csv file c:\Users\sorfanoudakis\Downloads\pv_netherlands.csv

data = pd.read_csv(r'pv_netherlands.csv', sep=',', header=0)
print(data)

#drop time column
data.drop('time', inplace=True, axis=1)
#rename local_time to time
data.rename(columns={'local_time':'time'}, inplace=True)

desired_timescale = 10 ##### from env
simulation_length = 96 ##### from env
simulation_date = '2023-2-1 08:00:00' ##### from env
number_of_transformers = 3 ##### from env

dataset_timescale = 60
dataset_starting_date = '2022-01-01 00:00:00'


if desired_timescale > dataset_timescale:
    data = data.groupby(data.index // (desired_timescale/dataset_timescale)).max()
elif desired_timescale < dataset_timescale:
    # extend the dataset to data.shape[0] * (dataset_timescale/desired_timescale)
    # by repeating the data every (dataset_timescale/desired_timescale) rows
    data = data.loc[data.index.repeat(dataset_timescale/desired_timescale)].reset_index(drop=True)
    # data = data/ (dataset_timescale/desired_timescale)

# smooth data by taking the mean of every 5 rows
data['electricity'] = data['electricity'].rolling(window=60//desired_timescale,min_periods=1).mean()
# use other type of smoothing
data['electricity'] = data['electricity'].ewm(span=60//desired_timescale, adjust=True).mean()

data['electricity'] = data['electricity'] * 100

#plot the data for the first 96 rows
data['electricity'][: (60//desired_timescale * 24*7) ].plot()

plt.show()




# duplicate the data to have two years of data
data = pd.concat([data, data], ignore_index=True)

#select only 2022

#plot in subplots

# plt.figure(figsize=(10, 7))
# for i,year in enumerate(range(2015,2024)):
#     data_temp = data[data['year'] == year]
#     plt.subplot(3, 3, i+1)
#     plt.title(f'year {year}')
#     plt.xlabel('hour')
#     plt.ylabel('price')
    # plt.plot(data_temp['Datetime (Local)'], data_temp['Price (EUR/MWhe)'])
    
# save_plots
# plt.savefig('data/eda_day-ahead_prices.png')



    
        







# #write script to read in data and do some basic EDA for the elaad data

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import os
import datetime

#read ev_specs json file
import json
with open('ev_specs.json') as f:
    data = json.load(f)
    
print(data.keys())

registrations = np.zeros(len(data.keys()))
for i, ev_name in enumerate(data.keys()):
    # sum the total number of registrations
    ev = data[ev_name]
    print(ev_name)
    registrations[i] = data[ev_name]['number_of_registrations_2023_nl']
print(registrations.sum())

normalized_registrations = registrations/registrations.sum()
#sample from the normalized registrations
sampled_ev = np.random.choice(list(data.keys()), p=normalized_registrations)
print(sampled_ev)
    
exit()

path =  "mean-session-length-per.csv"
df_time_of_stay_vs_arrival = pd.read_csv(path)
#replace Nan with 0
df_time_of_stay_vs_arrival = df_time_of_stay_vs_arrival.fillna(0)
print(df_time_of_stay_vs_arrival)
minute = 0
hour = 15
arrival_time = f'{hour:02d}:{minute:02d}'
print(arrival_time)
# print the mean energy demand for arrival time = hour:minute
print(df_time_of_stay_vs_arrival.loc[(df_time_of_stay_vs_arrival['Arrival Time'] == arrival_time)]['work'].values[0])



path =  "mean-demand-per-arrival.csv"
df_energy_demand = pd.read_csv(path)
#replace Nan with 0
df_energy_demand = df_energy_demand.fillna(0)
print(df_energy_demand)
minute = 0
hour = 15
arrival_time = f'{hour:02d}:{minute:02d}'
print(arrival_time)
# print the mean energy demand for arrival time = hour:minute
print(df_energy_demand.loc[(df_energy_demand['Arrival Time'] == arrival_time)]['work'].values[0])

exit()

# #print all file in directory
# for dirname, _, filenames in os.walk('.\data'):
#     for filename in filenames:
#         print(os.path.join(dirname, filename))

# # df_arrival_week = pd.read_csv('.\data\distribution-of-arrival.csv') #weekdays
# # df_arrival_weekend = pd.read_csv('.\data\distribution-of-arrival-weekend.csv') #weekends
# # df_connection_time = pd.read_csv('.\data\distribution-of-connection-time.csv') #connection time
# # df_denergy_demand = pd.read_csv('.\data\distribution-of-energy-demand.csv') #energy demand

# print(df_arrival_week['public'].sum())
# print(df_arrival_weekend['public'].sum())
# print(df_arrival_week)
# print(df_arrival_weekend)

# date = datetime.datetime(2021, 1, 1, 0, 0, 0)

# arrival_times = []
# energy_demand = []
# connection_time = []

# numbers = np.random.randint(0,100,size=1)

# for i in numbers:
#     energy_demand.append(df_denergy_demand['public'].iloc[i])
#     connection_time.append(df_connection_time['public'].iloc[i])
# plt.hist(energy_demand, bins=20)
# plt.show()
# plt.hist(connection_time, bins=24)
# for k in range(3650):
#     # for i in range(int(24*60/15)):    
#     #     date += datetime.timedelta(minutes=15)

#         # print(date)
#         #this is how I am gonna sample public ev spawns
#         #Allow to choose residential, public, workplace scenarios
#         # if np.random.rand(1)*100 < df_arrival_week['public'].iloc[i]:
#         #     arrival_times.append(i)

    
        
# # plot histogram of arrival times
# plt.hist(arrival_times, bins=96)
# plt.show()

# #read in data from xlsx file
df = pd.read_excel('.\data\elaadnl_open_ev_datasets.xlsx',
                    sheet_name='open_transactions')
#get all keys
keys = df.keys()

df['hour_of_arrival'] = df['UTCTransactionStart'].dt.hour

#drop all other columns except for "ChargeTime" and "UTCTransactionStart"
for k in keys:
    if k != "ChargeTime" and k != "hour_of_arrival":
        df = df.drop(k, axis=1)


print(df.head())

#plot a histogram of the hour_of_arrival vs the Chargetime for every hour in the same figure
# also add legends and labels, and make plots transparent
# plt.figure()
#save the bins and edges for later use

data_table = np.zeros((24,48))

for i in range(24):
    bins, edges = np.histogram(df['ChargeTime'].loc[df['hour_of_arrival'] == i], bins=48)
    #normalize the bins
    bins = bins/bins.sum()
    data_table[i,:] = bins

    #smooth the data
    bins = np.convolve(bins, np.ones(5)/5, mode='same')


    #plot cdf    
    # plt.plot(edges[:-1], np.cumsum(bins), label=f'Hour {i}')

    plt.plot(edges[:-1], bins, label=f'Hour {i}')
plt.legend()
plt.xlabel('Charge Time')
plt.ylabel('Frequency')
plt.title('Charge Time vs Hour of Arrival')
plt.show()

#save data_table as numpy array
np.save('time_of_connection_vs_hour.npy', data_table)



# This file contains the reward function for the RL agent
# Users can create their own reward function here or in their own file using the same structure as below

import math

def SquaredTrackingErrorReward(env,*args):
    # This reward function is the squared tracking error that uses the minimum of the power setpoints and the charge power potential
    # The reward is negative
    
    reward = - (min(env.power_setpoints[env.current_step-1], env.charge_power_potential[env.current_step-1]) -
        env.current_power_usage[env.current_step-1])**2
        
    return reward

def SqTrError_TrPenalty_UserIncentives(env, _, user_satisfaction_list, *args):
    # This reward function is the squared tracking error that uses the minimum of the power setpoints and the charge power potential
    # It penalizes transofrmers that are overloaded    
    # The reward is negative
    
    tr_max_limit = env.transformers[0].max_power[env.current_step-1]
    
    reward = - (min(env.power_setpoints[env.current_step-1], env.charge_power_potential[env.current_step-1],tr_max_limit) -
        env.current_power_usage[env.current_step-1])**2
            
    for tr in env.transformers:
        reward -= 100 * tr.get_how_overloaded()
        
    for score in user_satisfaction_list:
        reward -= 1000 * (1 - score)
                    
    return reward

def ProfitMax_TrPenalty_UserIncentives(env, total_costs, user_satisfaction_list, *args):
    # This reward function is the squared tracking error that uses the minimum of the power setpoints and the charge power potential
    # It penalizes transofrmers that are overloaded    
    # The reward is negative
    reward = total_costs
    
    for tr in env.transformers:
        reward -= 100 * tr.get_how_overloaded()                        
    
    for score in user_satisfaction_list:        
        reward -= 100 * math.exp(-10*score)
        
    return reward

def SquaredTrackingErrorRewardWithPenalty(env,*args):
    # This reward function is the squared tracking error that uses the minimum of the power setpoints and the charge power potential
    # The reward is negative
    # If the EV is not charging, the reward is penalized
    
    if env.current_power_usage[env.current_step-1] == 0 and env.charge_power_potential[env.current_step-2] != 0:
        reward = - (min(env.power_setpoints[env.current_step-1], env.charge_power_potential[env.current_step-1]) -
            env.current_power_usage[env.current_step-1])**2 - 100
    else:
        reward = - (min(env.power_setpoints[env.current_step-1], env.charge_power_potential[env.current_step-1]) -
            env.current_power_usage[env.current_step-1])**2
    
    return reward

def SimpleReward(env,*args):
    # This reward function does not consider the charge power potential
    
    reward = - (env.power_setpoints[env.current_step-1] - env.current_power_usage[env.current_step-1])**2
    
    return reward

def MinimizeTrackerSurplusWithChargeRewards(env,*args):
    # This reward function minimizes the tracker surplus and gives a reward for charging
    
    reward = 0
    if env.power_setpoints[env.current_step-1] < env.current_power_usage[env.current_step-1]:
            reward -= (env.current_power_usage[env.current_step-1]-env.power_setpoints[env.current_step-1])**2

    reward += env.current_power_usage[env.current_step-1] #/75
    
    return reward

def profit_maximization(env, total_costs, user_satisfaction_list, *args):
    # This reward function is the profit maximization reward function
    
    reward = total_costs
    
    for score in user_satisfaction_list:
        # reward -= 100 * (1 - score)
        reward -= 100 * math.exp(-10*score)
    
    return reward



# Previous reward functions for testing
#############################################################################################################
        # reward = total_costs  # - 0.5
        # print(f'total_costs: {total_costs}')
        # print(f'user_satisfaction_list: {user_satisfaction_list}')
        # for score in user_satisfaction_list:
        #     reward -= 100 * (1 - score)

        # Punish invalid actions (actions that try to charge or discharge when there is no EV connected)
        # reward -= 2 * (invalid_action_punishment/self.number_of_ports)

        # reward = min(2, 1 * 4 * self.cs / (0.00001 + (
        #     self.power_setpoints[self.current_step-1] - self.current_power_usage[self.current_step-1])**2))

        # this is the new reward function
        # reward = min(2, 1/((min(self.power_setpoints[self.current_step-1], self.charge_power_potential[self.current_step-1]) -
        #           self.current_power_usage[self.current_step-1])**2 + 0.000001))

        # new_*10*charging
        # if self.power_setpoints[self.current_step-1] < self.current_power_usage[self.current_step-1]:
        #     reward -= (min(self.power_setpoints[self.current_step-1], self.charge_power_potential[self.current_step-1]) -
        #             self.current_power_usage[self.current_step-1])
        # else:
        #     reward -= (min(self.power_setpoints[self.current_step-1], self.charge_power_potential[self.current_step-1]) -
        #             self.current_power_usage[self.current_step-1])*10

        # new_1_equal
        # if self.power_setpoints[self.current_step-1] < self.current_power_usage[self.current_step-1]:
        #     reward -= (min(self.power_setpoints[self.current_step-1], self.charge_power_potential[self.current_step-1]) -
        #             self.current_power_usage[self.current_step-1])
        # else:
        #     reward -= (min(self.power_setpoints[self.current_step-1], self.charge_power_potential[self.current_step-1]) -
        #             self.current_power_usage[self.current_step-1])

        # new_0.1
        # if self.power_setpoints[self.current_step-1] < self.current_power_usage[self.current_step-1]:
        #     reward -= (min(self.power_setpoints[self.current_step-1], self.charge_power_potential[self.current_step-1]) -
        #             self.current_power_usage[self.current_step-1])**2
        # else:
        #     reward -= (min(self.power_setpoints[self.current_step-1], self.charge_power_potential[self.current_step-1]) -
        #             self.current_power_usage[self.current_step-1])*0.1

        # new_reward squared
        # if self.power_setpoints[self.current_step-1] < self.current_power_usage[self.current_step-1]:
        #     reward -= (min(self.power_setpoints[self.current_step-1], self.charge_power_potential[self.current_step-1]) -
        #           self.current_power_usage[self.current_step-1])**2
        # else:
        #     reward -= (min(self.power_setpoints[self.current_step-1], self.charge_power_potential[self.current_step-1]) -
        #           self.current_power_usage[self.current_step-1])

        # for score in user_satisfaction_list:
        #     reward -= 100 * (1 - score)

        # for tr in self.transformers:
        #     if tr.current_amps > tr.max_current:
        #         reward -= 1000 * abs(tr.current_amps - tr.max_current)
        #     elif tr.current_amps < tr.min_current:
        #         reward -= 1000 * abs(tr.current_amps - tr.min_current)

        # reward -= 100 * (tr.current_amps < tr.min_amps)
        #######################################################################################################
        # squared tracking error
        # reward -= (min(self.power_setpoints[self.current_step-1], self.charge_power_potential[self.current_step-1]) -
        #            self.current_power_usage[self.current_step-1])**2

        # best reward so far
        ############################################################################################################
        # if self.power_setpoints[self.current_step-1] < self.current_power_usage[self.current_step-1]:
        #     reward -= (self.current_power_usage[self.current_step-1]-self.power_setpoints[self.current_step-1])

        # reward += self.current_power_usage[self.current_step-1]/75
        ############################################################################################################
        # normalize reward to -1 1
        # reward = reward/1000
        # reward = (100 +reward) / 1000
        # print(f'reward: {reward}')

        # reward -= 2 * (invalid_action_punishment/self.number_of_ports)
        # reward /= 100
        # reward = (100 +reward) / 1000
        # print(f'current_power_usage: {self.current_power_usage[self.current_step-1]}')

        # return reward


'''
===================================
Author: Stavros Orfanoudakis 2023
===================================
'''

import os
import numpy as np
import math
from ev2gym.utilities.utils import get_statistics

class EvCityReplay():
    '''
    This class is used to save the simulation data in a pickle file.
    The pickle file can be used to create a math model of the simulation.    
    '''

    def __init__(self, env):

        # Create replay folder if it does not exist
        if not os.path.exists('replay'):
            os.makedirs('replay')

        self.stats = get_statistics(env)

        self.replay_path = env.replay_path + 'replay_' + env.sim_name + '.pkl'
        self.sim_name = env.sim_name + '_replay'
        self.sim_length = env.simulation_length
        self.n_cs = env.cs
        self.n_transformers = env.number_of_transformers
        self.timescale = env.timescale
        self.sim_date = env.sim_starting_date
        # self.cs_transformers = env.cs_transformers
        self.power_setpoints = env.power_setpoints
        self.scenario = env.scenario
        self.heterogeneous_specs = env.heterogeneous_specs
        self.ev_load_potential = env.current_power_usage

        self.transformers = env.transformers
        self.charging_stations = env.charging_stations
        self.EVs = env.EVs
        
        self.unstirred_EVs = None
        self.unstirred_stats = None
        self.optimal_EVs = None
        self.optimal_stats = None
        
        #save unstirred and optimal EVs and stats for evaluation
        if env.eval_mode =="unstirred":
            self.unstirred_EVs = env.EVs
            self.unstirred_stats = self.stats
        
        elif env.eval_mode =="optimal":
            self.optimal_EVs = env.EVs
            self.optimal_stats = self.stats
            
            if env.load_from_replay_path is not None:            
                self.unstirred_EVs = env.replay.unstirred_EVs
                self.unstirred_stats = env.replay.unstirred_stats
        else:
            if env.load_from_replay_path is not None:            
                self.unstirred_EVs = env.replay.unstirred_EVs
                self.unstirred_stats = env.replay.unstirred_stats                    
                self.optimal_EVs = env.replay.optimal_EVs
                self.optimal_stats = env.replay.optimal_stats            
    
        
        # self.transformer_amps  = env.transformer_amps
        # self.cs_power = env.cs_power
        # self.port_power = env.port_power
        
        self.simulate_grid = env.simulate_grid

        self.charge_prices = env.charge_prices
        self.discharge_prices = env.discharge_prices

        self.tra_max_amps = np.ones([self.n_transformers, self.sim_length])
        self.tra_min_amps = np.ones([self.n_transformers, self.sim_length])

        for i, tra in enumerate(env.transformers):
            current_from_inflexible = env.tr_inflexible_loads[i,:] * 1000 / 400
            current_from_solar = env.tr_solar_power[i,:] * 1000 / 400
            
            self.tra_max_amps[i] = tra.max_current - abs(current_from_inflexible) + abs(current_from_solar)
            self.tra_min_amps[i] = tra.min_current + abs(current_from_inflexible) - abs(current_from_solar)

        self.port_max_charge_current = np.ones([self.n_cs])
        self.port_min_charge_current = np.ones([self.n_cs])
        self.port_max_discharge_current = np.ones([self.n_cs])
        self.port_min_discharge_current = np.ones([self.n_cs])
        self.voltages = np.ones([self.n_cs])
        self.phases = np.ones([self.n_cs])        

        self.cs_ch_efficiency = np.ones([self.n_cs, self.sim_length])
        self.cs_dis_efficiency = np.ones([self.n_cs, self.sim_length])
        self.cs_transformer = np.ones([self.n_cs])

        self.max_n_ports = 0

        for i, cs in enumerate(env.charging_stations):
            self.port_max_charge_current[i] = cs.max_charge_current
            self.port_min_charge_current[i] = cs.min_charge_current
            self.port_max_discharge_current[i] = cs.max_discharge_current
            self.port_min_discharge_current[i] = cs.min_discharge_current
            self.voltages[i] = cs.voltage * math.sqrt(cs.phases)            

            #consider usecases with variable number of ports per cs
            if cs.n_ports > self.max_n_ports:
                self.max_n_ports = cs.n_ports

            self.cs_transformer[i] = cs.connected_transformer

        self.ev_max_energy = np.zeros([self.max_n_ports,
                                      self.n_cs,
                                      self.sim_length])  # ev max battery capacity, 0 if no ev is there
        self.ev_min_energy = np.zeros([self.max_n_ports,
                                       self.n_cs,
                                       self.sim_length])  # ev min battery capacity, 0 if no ev is there
        self.ev_max_ch_power = np.zeros([self.max_n_ports,
                                         self.n_cs,
                                         self.sim_length])  # ev max charging power, 0 if no ev is there
        # self.ev_min_ch_power = np.zeros([self.max_n_ports,
        #                                  self.n_cs,
        #                                  self.sim_length])  # ev min charging power, 0 if no ev is there
        self.ev_max_dis_power = np.zeros([self.max_n_ports,
                                          self.n_cs,
                                          self.sim_length])  # ev max discharging power, 0 if no ev is there
        # self.ev_min_dis_power = np.zeros([self.max_n_ports,
        #                                   self.n_cs,
        #                                   self.sim_length])  # ev min discharging power, 0 if no ev is there
        self.u = np.zeros([self.max_n_ports,
                           self.n_cs,
                           self.sim_length])  # u is 0 if port is empty and 1 if port is occupied
        self.energy_at_arrival = np.zeros([self.max_n_ports,
                                           self.n_cs,
                                           self.sim_length])  # x when ev arrives at the port
        self.ev_arrival = np.zeros([self.max_n_ports,
                                    self.n_cs,
                                    self.sim_length])  # 1 when an ev arrives-> power = 0 and energy = x
        self.t_dep = np.zeros([self.max_n_ports,
                               self.n_cs,
                               self.sim_length])  # time of departure of the ev, 0 if port is empty
        self.ev_des_energy = np.zeros([self.max_n_ports,
                                       self.n_cs,
                                       self.sim_length])  # desired energy of the ev, 0 if port is empty
        self.max_energy_at_departure = np.zeros([self.max_n_ports,
                                                self.n_cs,
                                                self.sim_length])  # max energy of ev when only charging

        for i, ev in enumerate(env.EVs):
            port = ev.id
            cs_id = ev.location
            t_arr = ev.time_of_arrival
            original_t_dep = ev.time_of_departure
            # print(f'EV {i} is at port {port} of CS {cs_id} from {t_arr} to {original_t_dep}')            

            if t_arr >= self.sim_length:
                continue
            if original_t_dep >= self.sim_length:
                t_dep = self.sim_length
            else:
                t_dep = original_t_dep                            

            self.ev_max_energy[port, cs_id, t_arr:t_dep] = ev.battery_capacity
            # self.ev_min_energy[port, cs_id, t_arr:t_dep] = ev.battery_capacity * ev.min_soc
            self.ev_max_ch_power[port, cs_id,
                                 t_arr:t_dep] = ev.max_ac_charge_power
            # self.ev_min_ch_power[port, cs_id,
            #                      t_arr:t_dep] = 0
            self.ev_max_dis_power[port, cs_id,
                                  t_arr:t_dep] = ev.max_discharge_power
            # self.ev_min_dis_power[port, cs_id,
            #                       t_arr:t_dep] = 0
            self.u[port, cs_id, t_arr:t_dep] = 1            
            self.energy_at_arrival[port, cs_id,
                                   t_arr] = ev.battery_capacity_at_arrival
            self.ev_arrival[port, cs_id, t_arr] = 1
            if original_t_dep < self.sim_length:
                self.t_dep[port, cs_id, t_dep] = 1            
                if ev.prev_capacity < ev.battery_capacity:
                    self.max_energy_at_departure[port, cs_id, t_dep] = ev.prev_capacity #-5
                else:
                    self.max_energy_at_departure[port, cs_id, t_dep] = ev.battery_capacity
            else:
                self.t_dep[port, cs_id, t_dep-1] = 1                            
                self.max_energy_at_departure[port, cs_id, t_dep-1] = ev.prev_capacity
            
            self.ev_des_energy[port, cs_id, t_dep] = ev.desired_capacity

        # print(f'u: {self.u}')
        # print(f'ev_arrival: {self.ev_arrival}')
        # print(f't_dep: {self.t_dep}')
        # print(f'ev_des_energy: {self.ev_des_energy}')
        # print(f'ev_max_energy: {self.ev_max_energy}')
        # print(f'ev_max_ch_power: {self.ev_max_ch_power}')
        # print(f'ev_max_dis_power: {self.ev_max_dis_power}')



import matplotlib.pyplot as plt
import networkx as nx
import PIL
import numpy as np
import pkg_resources

# Image URLs for graph nodes

icons = {
    "cpo": pkg_resources.resource_filename('ev2gym', "visuals/icons/cpo.png"),
    "transformer": pkg_resources.resource_filename('ev2gym', "visuals/icons/transformer.png"),
    "charger_1": pkg_resources.resource_filename('ev2gym', "visuals/icons/charging-station_1_port.png"),
    "charger_2": pkg_resources.resource_filename('ev2gym', "visuals/icons/charging-station_2_ports.png"),
    "charger_wallbox": pkg_resources.resource_filename('ev2gym', "visuals/icons/charger_wallbox.png"),
    "ev": pkg_resources.resource_filename('ev2gym', "visuals/icons/ev.png")
}


class Renderer():
    """Class for rendering the simulation environment"""

    def __init__(self, env):
        self.env = env

        # create networkx graph from environment
        self.G = nx.Graph()

        # Load images
        images = {k: PIL.Image.open(fname) for k, fname in icons.items()}

        # add nodes
        self.G.add_node("cpo", image=images["cpo"])

        for i in range(0, self.env.number_of_transformers):
            self.G.add_node(f"transformer_{i}", image=images["transformer"])
            # add edge to cpo
            self.G.add_edge("cpo", "transformer_" + str(i))

        for i in range(0, self.env.cs):
            if self.env.charging_stations[i].n_ports == 1:
                self.G.add_node("charger_" + str(i), image=images["charger_1"])
            elif self.env.charging_stations[i].n_ports == 2:
                self.G.add_node("charger_" + str(i), image=images["charger_2"])
            else:
                self.G.add_node("charger_" + str(i), image=images["charger_2"])

            # add edge to transformer
            self.G.add_edge(
                "transformer_" + str(self.env.charging_stations[i].connected_transformer), "charger_" + str(i))

            for j in range(0, self.env.charging_stations[i].n_ports):
                # add ev node
                self.G.add_node("ev_" + str(i) + "_" +
                                str(j), image=images["ev"])
                # add edge to charger
                self.G.add_edge("charger_" + str(i), "ev_" +
                                str(i) + "_" + str(j))

        # Get a reproducible layout and create figure
        pos = nx.spring_layout(self.G, seed=42)
        pos = nx.kamada_kawai_layout(self.G, pos=pos, weight=None)
        pos_attrs = {}
        for node, coords in pos.items():
            pos_attrs[node] = (coords[0], coords[1] + 0.14)

        pos__down_attrs = {}
        for node, coords in pos.items():
            pos__down_attrs[node] = (coords[0], coords[1] - 0.14)

        plt.ion()
        self.fig, ax = plt.subplots()

        # Note: the min_source/target_margin kwargs only work with FancyArrowPatch objects.
        # Force the use of FancyArrowPatch for edge drawing by setting `arrows=True`,
        # but suppress arrowheads with `arrowstyle="-"`
        self.graph_edges = nx.draw_networkx_edges(
            self.G,
            pos=pos,
            ax=ax,
            arrows=True,
            arrowstyle="-",
            min_source_margin=15,
            min_target_margin=15,
        )
    
        #change the linestype of the edges connectign transformer to cpo
        for i in range(0, self.env.number_of_transformers):
            self.graph_edges[i].set_linestyle("--")
            self.graph_edges[i].set_linewidth(3)
            

        # Drawnx labels for each charging station node to show power

        node_labels = {}
        for i in range(0, self.env.cs):
            node_labels["charger_" +
                        str(i)] = str(round(self.env.charging_stations[i].current_power_output, 1)) + "kW"

        self.charger_labels_ax = nx.draw_networkx_labels(self.G,
                                                         pos=pos__down_attrs,
                                                         labels=node_labels,
                                                         ax=ax,
                                                         verticalalignment='bottom',
                                                         font_size=8,
                                                         font_color="blue"
                                                         )

        # Draw charger current
        # draw nx labels for each edge from charger to transformer

        edge_labels = {}
        for i in range(0, self.env.cs):
            for j in range(0, self.env.charging_stations[i].n_ports):
                edge_labels[("charger_" + str(i), "transformer_" + str(self.env.charging_stations[i].connected_transformer))] = str(
                    round(self.env.charging_stations[i].current_total_amps, 1)) + "A"

        self.charger_to_tr_axes = nx.draw_networkx_edge_labels(self.G,
                                                               pos=pos,
                                                               edge_labels=edge_labels,
                                                               ax=ax,
                                                               font_color="r",
                                                               )

        # Draw node label cpo total power consumed

        self.cpo_power_ax = nx.draw_networkx_labels(self.G,
                                                    pos=pos__down_attrs,
                                                    labels={"cpo": str(
                                                    round(self.env.current_power_usage[self.env.current_step], 1)) + "kW"},
                                                    ax=ax,
                                                    verticalalignment='bottom',
                                                    font_size=12,
                                                    font_color="blue"
                                                    )

        # Draw consumed power for each transformer
        # draw node labels for each transformer
        node_labels = {}
        for i in range(0, self.env.number_of_transformers):
            # append transformer current to node labels
            node_labels["transformer_" +
                        str(i)] = str(round(self.env.transformers[i].current_amps, 2)) + "A"

        self.tr_labels_ax = nx.draw_networkx_labels(self.G,
                                                    pos=pos_attrs,
                                                    labels=node_labels,
                                                    ax=ax,
                                                    verticalalignment='top',
                                                    font_color="orange"
                                                    )

        # draw edge labels for each ev to charger
        edge_labels = {}
        for i in range(0, self.env.cs):
            for j in range(0, self.env.charging_stations[i].n_ports):
                if self.env.charging_stations[i].evs_connected[j] != None:
                    edge_labels[("ev_" + str(i) + "_" + str(j), "charger_" + str(i))] = str(
                        round(self.env.charging_stations[i].evs_connected[j].actual_current, 1)) + "A"
                else:
                    edge_labels[("ev_" + str(i) + "_" + str(j),
                                 "charger_" + str(i))] = ""

        self.ev_to_charger_labels_ax = nx.draw_networkx_edge_labels(self.G,
                                                                    pos=pos,
                                                                    edge_labels=edge_labels,
                                                                    ax=ax,
                                                                    font_color="r",
                                                                    font_size=8,
                                                                    )

        # Add label to the cpo node
        nx.draw_networkx_labels(self.G,
                                pos=pos_attrs,
                                labels={"cpo": "CPO"},
                                ax=ax,
                                verticalalignment='top',
                                font_color="black"
                                )

        # Transform from data coordinates (scaled between xlim and ylim) to display coordinates
        tr_figure = ax.transData.transform
        # Transform from display to figure coordinates
        tr_axes = self.fig.transFigure.inverted().transform

        # Select the size of the image (relative to the X axis)
        icon_size = (ax.get_xlim()[1] - ax.get_xlim()[0]) * 0.025
        icon_center = icon_size / 2.0

        self.ev_list = []
        # Add the respective image to each node
        for n in self.G.nodes:
            xf, yf = tr_figure(pos[n])
            xa, ya = tr_axes((xf, yf))
            # get overlapped axes and plot icon
            a = plt.axes(
                [xa - icon_center, ya - icon_center, icon_size, icon_size])
            a.imshow(self.G.nodes[n]["image"])
            a.axis("off")

            if n.startswith("ev"):
                self.ev_list.append(a)

        # add title to top of the figure
        # self.fig.suptitle("EV Charging Network", fontsize=16)

        # add icon to the bottom of the figure
        image_path = pkg_resources.resource_filename(
            'ev2gym', "visuals/icons/logo.png")
        icon = PIL.Image.open(image_path)
        icon = icon.resize((100, 50), PIL.Image.LANCZOS)
        self.fig.figimage(icon, 20, 20, alpha=1, zorder=1)

        # add text to a box in the bottom right corner
        text = f"Time step: {self.env.current_step}/ {self.env.simulation_length}\n" + \
               f"{self.env.sim_date}\n" \
            f"Simulation Name: {self.env.sim_name}"

        self.text_box = self.fig.text(
            0.99,
            0.01,
            text,
            wrap=True,
            horizontalalignment="right",
            fontsize=12,
            bbox={"facecolor": "white", "alpha": 0.5, "pad": 5},
        )

        # make ev nodes invisible
        counter = 0
        for i in range(0, self.env.cs):
            for j in range(0, self.env.charging_stations[i].n_ports):

                if self.env.charging_stations[i].evs_connected[j] == None:
                    self.ev_list[counter].set_visible(False)

                counter += 1

        self.fig.canvas.draw()
        self.fig.canvas.flush_events()

    def render(self):
        # input("Press Enter to continue...")
        text = f"Time step: {self.env.current_step} / {self.env.simulation_length}\n" + \
               f"{self.env.sim_date}\n" \
            f"Scenario: {self.env.scenario}\n" \
            f"Simulation Name: {self.env.sim_name}\n"

        self.text_box.set_text(text)
        self.cpo_power_ax["cpo"].set_text(str(
            round(self.env.current_power_usage[self.env.current_step-1], 1)) + "kW")

        counter = 0
        for i in range(0, self.env.cs):
            for j in range(0, self.env.charging_stations[i].n_ports):
                if self.env.charging_stations[i].evs_connected[j] == None:
                    self.ev_list[counter].set_visible(False)
                    self.ev_to_charger_labels_ax[(
                        "ev_" + str(i) + "_" + str(j), "charger_" + str(i))].set_text("")
                else:
                    self.ev_list[counter].set_visible(True)
                    # set text to show soc
                    self.ev_list[counter].set_title(
                        str(round(
                            self.env.charging_stations[i].evs_connected[j].get_soc(), 1)*100) + "%",
                        fontsize=8,
                        color="green",)

                    self.ev_to_charger_labels_ax[("ev_" + str(i) + "_" + str(j), "charger_" + str(i))].set_text(str(
                        round(self.env.charging_stations[i].evs_connected[j].actual_current, 1)) + "A")

                counter += 1

            self.charger_to_tr_axes[("charger_" + str(i), "transformer_" + str(self.env.charging_stations[i].connected_transformer))
                                    ].set_text(str(round(self.env.charging_stations[i].current_total_amps, 1)) + "A")

            self.charger_labels_ax["charger_" + str(i)].set_text(
                str(round(self.env.charging_stations[i].current_power_output, 1)) + "kW")

        for i in range(0, self.env.number_of_transformers):
            # append transformer current to node labels
            self.tr_labels_ax["transformer_" +
                              str(i)].set_text(str(round(self.env.transformers[i].current_amps /
                                                         self.env.transformers[i].max_current[self.env.current_step-1]
                                                         *100, 0)) + "%")
        # update ev nodes
        self.fig.canvas.draw()
        self.fig.canvas.flush_events()



# This file contains support functions for the EV City environment.

import numpy as np
import matplotlib.pyplot as plt
import math
import datetime
from typing import List, Dict

from ev2gym.models.ev import EV


def get_statistics(env) -> Dict:
    total_ev_served = np.array(
        [cs.total_evs_served for cs in env.charging_stations]).sum()
    total_profits = np.array(
        [cs.total_profits for cs in env.charging_stations]).sum()
    total_energy_charged = np.array(
        [cs.total_energy_charged for cs in env.charging_stations]).sum()
    total_energy_discharged = np.array(
        [cs.total_energy_discharged for cs in env.charging_stations]).sum()
    average_user_satisfaction = np.array(
        [cs.get_avg_user_satisfaction() for cs in env.charging_stations
         if cs.total_evs_served > 0]).mean()
    # get transformer overload from env.tr_overload
    total_transformer_overload = np.array(env.tr_overload).sum()

    tracking_error = 0
    energy_tracking_error = 0
    power_tracker_violation = 0
    for t in range(env.simulation_length):
        # tracking_error += (min(env.power_setpoints[t], env.charge_power_potential[t]) -
        #                    env.current_power_usage[t])**2
        # energy_tracking_error += abs(min(env.power_setpoints[t], env.charge_power_potential[t]) -
        #                              env.current_power_usage[t])

        tracking_error += (env.power_setpoints[t] -
                           env.current_power_usage[t])**2
        energy_tracking_error += abs(env.power_setpoints[t] -
                                     env.current_power_usage[t])

        if env.current_power_usage[t] > env.power_setpoints[t]:
            power_tracker_violation += env.current_power_usage[t] - \
                env.power_setpoints[t]

    energy_tracking_error *= env.timescale / 60

    # calculate total batery degradation
    battery_degradation = np.array(
        [np.array(ev.get_battery_degradation()) for ev in env.EVs])
    if len(battery_degradation) == 0:
        battery_degradation = np.zeros((1, 2))
    battery_degradation_calendar = battery_degradation[:, 0].sum()
    battery_degradation_cycling = battery_degradation[:, 1].sum()
    battery_degradation = battery_degradation.sum()

    # find the final battery capacity of evs
    if env.eval_mode != "unstirred" and len(env.EVs) > 0 \
            and env.replay is not None:

        if env.replay.unstirred_EVs is None:
            energy_user_satisfaction = np.array(-10000000)
        else:
            energy_user_satisfaction = np.zeros((len(env.EVs)))
            for i, ev in enumerate(env.EVs):
                e_actual = ev.current_capacity
                e_max = env.replay.unstirred_EVs[i].current_capacity
                # print(f'EV {i} actual: {e_actual:.2f} kWh, max: {e_max:.2f} kWh')
                energy_user_satisfaction[i] = e_actual / e_max * 100

        energy_user_satisfaction = energy_user_satisfaction.mean()
    else:
        energy_user_satisfaction = 100

    stats = {'total_ev_served': total_ev_served,
             'total_profits': total_profits,
             'total_energy_charged': total_energy_charged,
             'total_energy_discharged': total_energy_discharged,
             'average_user_satisfaction': average_user_satisfaction,
             'power_tracker_violation': power_tracker_violation,
             'tracking_error': tracking_error,
             'energy_tracking_error': energy_tracking_error,
             'energy_user_satisfaction': energy_user_satisfaction,
             'total_transformer_overload': total_transformer_overload,
             'battery_degradation': battery_degradation,
             'battery_degradation_calendar': battery_degradation_calendar,
             'battery_degradation_cycling': battery_degradation_cycling,
             'total_reward': env.total_reward,
             }

    if env.eval_mode != "optimal" and env.replay is not None:
        if env.replay.optimal_stats is not None:
            stats['opt_profits'] = env.replay.optimal_stats["total_profits"]
            stats['opt_tracking_error'] = env.replay.optimal_stats["tracking_error"]
            stats['opt_actual_tracking_error'] = env.replay.optimal_stats["energy_tracking_error"]
            stats['opt_power_tracker_violation'] = env.replay.optimal_stats["power_tracker_violation"]
            stats['opt_energy_user_satisfaction'] = env.replay.optimal_stats["energy_user_satisfaction"]
            stats['opt_total_energy_charged'] = env.replay.optimal_stats["total_energy_charged"]

    return stats


def print_statistics(env) -> None:

    stats = get_statistics(env)

    total_ev_served = stats['total_ev_served']
    total_profits = stats['total_profits']
    total_energy_charged = stats['total_energy_charged']
    total_energy_discharged = stats['total_energy_discharged']
    average_user_satisfaction = stats['average_user_satisfaction']
    total_transformer_overload = stats['total_transformer_overload']
    tracking_error = stats['tracking_error']
    energy_tracking_error = stats['energy_tracking_error']
    power_tracker_violation = stats['power_tracker_violation']
    energy_user_satisfaction = stats['energy_user_satisfaction']
    total_transformer_overload = stats['total_transformer_overload']
    battery_degradation = stats['battery_degradation']
    battery_degradation_calendar = stats['battery_degradation_calendar']
    battery_degradation_cycling = stats['battery_degradation_cycling']

    print("\n\n==============================================================")
    print("Simulation statistics:")
    for cs in env.charging_stations:
        print(cs)
    print(
        f'  - Total EVs spawned: {env.total_evs_spawned} |  served: {total_ev_served}')
    print(f'  - Total profits: {total_profits:.2f} €')
    print(
        f'  - Average user satisfaction: {average_user_satisfaction*100:.2f} %')

    print(
        f'  - Total energy charged: {total_energy_charged:.1f} | discharged: {total_energy_discharged:.1f} kWh')
    print(
        f'  - Power Tracking squared error: {tracking_error:.2f}, Power Violation: {power_tracker_violation:.2f} kW')
    print(f' - Actual Energy Tracking error: {energy_tracking_error:.2f} kW')
    print(f'  - Energy user satisfaction: {energy_user_satisfaction:.2f} %')
    print(
        f'  - Total Battery degradation: {battery_degradation:.5f}% | Calendar: {battery_degradation_calendar:.5f}%, Cycling: {battery_degradation_cycling:.5f}%')
    print(
        f'  - Total transformer overload: {total_transformer_overload:.2f} kWh \n')

    print("==============================================================\n\n")


def spawn_single_EV(env,
                    scenario,
                    cs_id,
                    port,
                    hour,
                    minute,
                    step,
                    min_time_of_stay_steps
                    ) -> EV:
    '''
    This function spawns a single EV and returns it
    '''

    # required energy independent of time of arrival
    # required_energy = env.df_energy_demand[scenario].iloc[np.random.randint(
    #     0, 100, size=1)].values[0]  # kWh

    # roound minute to 30 or 0
    if minute < 30:
        minute = 0
    else:
        minute = 30

    # required energy dependent on time of arrival
    arrival_time = f'{hour:02d}:{minute:02d}'

    required_energy_mean = env.df_req_energy[
        (env.df_req_energy['Arrival Time'] == arrival_time)
    ][scenario].values[0]

    required_energy = np.random.normal(
        required_energy_mean, 0.5*required_energy_mean)  # kWh

    if required_energy < 5:
        required_energy = np.random.randint(5, 10)

    if env.heterogeneous_specs:
        sampled_ev = np.random.choice(
            list(env.ev_specs.keys()), p=env.normalized_ev_registrations)
        battery_capacity = env.ev_specs[sampled_ev]["battery_capacity"]
    else:
        battery_capacity = env.config["ev"]["battery_capacity"]

    if battery_capacity < required_energy:
        initial_battery_capacity = np.random.randint(1, battery_capacity)
    else:
        initial_battery_capacity = battery_capacity - required_energy
        
    if initial_battery_capacity > env.config["ev"]['desired_capacity']:
        initial_battery_capacity = np.random.randint(1, battery_capacity)
        
    if initial_battery_capacity < env.config["ev"]['min_battery_capacity']:
        initial_battery_capacity = env.config["ev"]['min_battery_capacity']        

    # time of stay dependent on time of arrival
    time_of_stay_mean = env.df_time_of_stay_vs_arrival[(
        env.df_time_of_stay_vs_arrival['Arrival Time'] == arrival_time)
    ][scenario].values[0]

    time_of_stay = np.random.normal(
        time_of_stay_mean, 0.2*time_of_stay_mean)  # hours

    # turn from hours to steps
    time_of_stay = time_of_stay * 60 / env.timescale + 1

    # Alternative method for time of stay based on 10.000 charging sessions
    # time_of_stay = np.random.choice(
    #     np.arange(0, 48, 1), 1, p=env.time_of_connection_vs_hour[hour, :])/2
    # time_of_stay = time_of_stay[0] * 60 / env.timescale + 1

    # Alternative method for time of stay without taking into account the hour
    # time_of_stay = env.df_connection_time[scenario].iloc[np.random.randint(
    #     0, 100, size=1)].values[0] * 60 / env.timescale + 1

    if time_of_stay < min_time_of_stay_steps:
        time_of_stay = min_time_of_stay_steps

    if env.empty_ports_at_end_of_simulation:        
        if time_of_stay + step + 4 >= env.simulation_length:
            return None
            time_of_stay = env.simulation_length - step - 4 - 2

    if env.heterogeneous_specs:
        return EV(id=port,
                  location=cs_id,
                  battery_capacity_at_arrival=initial_battery_capacity,
                  max_ac_charge_power=env.ev_specs[sampled_ev]["max_ac_charge_power"],
                  max_dc_charge_power=env.ev_specs[sampled_ev]["max_dc_charge_power"],
                  max_discharge_power=-env.ev_specs[sampled_ev]["max_dc_discharge_power"],
                  discharge_efficiency=np.round(1 -
                                                (np.random.rand()+0.00001)/20, 3),  # [0.95-1]
                  transition_soc=np.round(0.9 - \
                                          (np.random.rand()+0.00001)/5, 3),  # [0.7-0.9]
                  battery_capacity=battery_capacity,
                  desired_capacity=0.8*battery_capacity,
                  time_of_arrival=step+1,
                  time_of_departure=int(
                      time_of_stay + step + 3),
                  ev_phases=3,
                  timescale=env.timescale,
                  )
    else:
        return EV(id=port,
                  location=cs_id,
                  battery_capacity_at_arrival=initial_battery_capacity,
                  battery_capacity=battery_capacity,
                  desired_capacity=env.config["ev"]['desired_capacity'],
                  max_ac_charge_power=env.config["ev"]['max_ac_charge_power'],
                  min_ac_charge_power=env.config["ev"]['min_ac_charge_power'],
                  max_dc_charge_power=env.config["ev"]['max_dc_charge_power'],
                  max_discharge_power=env.config["ev"]['max_discharge_power'],
                  min_discharge_power=env.config["ev"]['min_discharge_power'],
                  time_of_arrival=step+1,
                  time_of_departure=int(
                      time_of_stay + step + 3),
                  ev_phases=env.config["ev"]['ev_phases'],
                  transition_soc=env.config["ev"]['transition_soc'],
                  charge_efficiency=env.config["ev"]['charge_efficiency'],
                  discharge_efficiency=env.config["ev"]['discharge_efficiency'],
                  timescale=env.timescale,
                  )


def EV_spawner(env) -> List[EV]:
    '''
    This function spawns all the EVs of the current simulation and returns the list of EVs

    Returns:
        EVs: list of EVs
    '''

    ev_list = []

    occupancy_list = np.zeros((env.number_of_ports, env.simulation_length))

    arrival_probabilities = np.random.rand(
        env.number_of_ports, env.simulation_length)

    scenario = env.scenario
    user_spawn_multiplier = env.config["spawn_multiplier"]
    time = env.sim_date

    # Define minimum time of stay duration so that an EV can fully charge
    min_time_of_stay = env.config['ev']["min_time_of_stay"]
    min_time_of_stay_steps = min_time_of_stay // env.timescale

    for t in range(2, env.simulation_length-min_time_of_stay_steps-1):
        day = time.weekday()
        hour = time.hour
        minute = time.minute
        # Divide by 15 because the spawn rate is in 15 minute intervals (in the csv file)
        i = hour*4 + minute//15

        if day < 5:
            if scenario == "workplace" and (hour < 6 or hour > 18):
                time = time + datetime.timedelta(minutes=env.timescale)
                continue
            else:
                tau = env.df_arrival_week[scenario].iloc[i]
                multiplier = 1  # 10
        else:
            if scenario == "workplace":
                time = time + datetime.timedelta(minutes=env.timescale)
                continue
            else:
                tau = env.df_arrival_weekend[scenario].iloc[i]

            if day == 5:
                multiplier = 1  # 8
            else:
                multiplier = 1  # 6

        counter = 0
        for cs in env.charging_stations:
            for port in range(cs.n_ports):
                # if port is empty
                if occupancy_list[counter, t] == 0 and \
                    occupancy_list[counter, t-1] == 0 and \
                        occupancy_list[counter, t-2] == 0:
                    # and there is an EV arriving
                    if arrival_probabilities[counter, t]*100 < tau * multiplier * (env.timescale/60) * user_spawn_multiplier:
                        ev = spawn_single_EV(
                            env=env,
                            scenario=scenario,
                            cs_id=cs.id,
                            port=port,
                            hour=hour,
                            minute=minute,
                            step=t,
                            min_time_of_stay_steps=min_time_of_stay_steps)

                        if ev is not None:                        
                            ev_list.append(ev)
                            occupancy_list[counter, t +
                                        1:ev.time_of_departure] = 1

                counter += 1
        # step the time
        time = time + datetime.timedelta(minutes=env.timescale)

    return ev_list


def smooth_vector(v) -> np.ndarray:
    n = len(v)
    smoothed_v = [0] * n

    # Calculate the sum of the original vector
    total_sum = sum(v)

    for i in range(n):
        # Calculate the range for averaging
        start = max(0, i - 1)
        end = min(n, i + 2)

        # Calculate the average of neighboring elements
        smoothed_v[i] = sum(v[start:end]) / (end - start)

    # Adjust the smoothed vector to maintain the original sum
    smoothed_sum = sum(smoothed_v)
    sum_ratio = total_sum / smoothed_sum

    # Apply the ratio to each element of the smoothed vector
    smoothed_v = [value * sum_ratio for value in smoothed_v]

    return smoothed_v


def median_smoothing(v, window_size) -> np.ndarray:
    smoothed_v = np.zeros_like(v)
    half_window = window_size // 2

    for i in range(len(v)):
        start = max(0, i - half_window)
        end = min(len(v), i + half_window + 1)
        smoothed_v[i] = np.median(v[start:end])

    return smoothed_v


def generate_power_setpoints(env) -> np.ndarray:
    '''
    This function generates the power setpoints for the entire simulation using
    the list of EVs and the charging stations from the environment.

    It considers the ev SoC and teh steps required to fully charge the EVs.

    Returns:
        power_setpoints: np.ndarray

    '''

    power_setpoints = np.zeros(env.simulation_length)
    # get normalized prices
    prices = abs(env.charge_prices[0])
    prices = prices / np.max(prices)

    required_energy_multiplier = 100 + \
        env.config["power_setpoint_flexiblity"]

    min_cs_power = env.charging_stations[0].get_min_charge_power()
    max_cs_power = env.charging_stations[0].get_max_power()

    total_evs_spawned = 0
    for t in range(env.simulation_length):
        counter = total_evs_spawned
        for _, ev in enumerate(env.EVs_profiles[counter:]):
            if ev.time_of_arrival == t + 1:
                total_evs_spawned += 1

                required_energy = ev.battery_capacity - ev.battery_capacity_at_arrival
                required_energy *= required_energy_multiplier / 100
                min_power_limit = max(ev.min_ac_charge_power, min_cs_power)
                max_power_limit = min(ev.max_ac_charge_power, max_cs_power)

                # Spread randomly the required energy over the time of stay using the prices as weights
                shifted_load = np.random.normal(loc=1 - prices[t+2:ev.time_of_departure],
                                                scale=min(
                                                    prices[t+2:ev.time_of_departure]),
                                                size=ev.time_of_departure - t - 2)
                # make shifted load positive
                shifted_load = np.abs(shifted_load)
                shifted_load = shifted_load / np.sum(shifted_load)
                shifted_load = shifted_load * required_energy * 60 / env.timescale

                # find power lower than min_power_limit and higher than max_power_limit
                step = 0
                while np.min(shifted_load[shifted_load != 0]) < min_power_limit or \
                        np.max(shifted_load) > max_power_limit:

                    if step > 10:
                        break

                    # print(f"Shifted load: {shifted_load}")
                    for i in range(len(shifted_load)):
                        if shifted_load[i] < min_power_limit and shifted_load[i] > 0:
                            load_to_shift = shifted_load[i]
                            shifted_load[i] = 0

                            if i == len(shifted_load) - 1:
                                shifted_load[0] += load_to_shift
                            else:
                                shifted_load[i+1] += load_to_shift

                        elif shifted_load[i] > max_power_limit:
                            load_to_shift = shifted_load[i] - max_power_limit
                            shifted_load[i] = max_power_limit

                            if i == len(shifted_load) - 1:
                                shifted_load[0] += load_to_shift
                            else:
                                shifted_load[i+1] += load_to_shift
                    step += 1

                power_setpoints[t+2:ev.time_of_departure] += shifted_load

            elif ev.time_of_arrival > t + 1:
                break

    # return smooth_vector(power_setpoints)

    return median_smoothing(power_setpoints, 5)


def calculate_charge_power_potential(env) -> float:
    '''
    This function calculates the total charge power potential of all currently parked EVs for the current time step     
    '''

    power_potential = 0
    for cs in env.charging_stations:
        cs_power_potential = 0
        for port in range(cs.n_ports):
            ev = cs.evs_connected[port]
            if ev is not None:
                if ev.get_soc() < 1 and ev.time_of_departure > env.current_step:
                    phases = min(cs.phases, ev.ev_phases)
                    ev_current = ev.max_ac_charge_power * \
                        1000/(math.sqrt(phases)*cs.voltage)
                    current = min(cs.max_charge_current, ev_current)
                    cs_power_potential += math.sqrt(phases) * \
                        cs.voltage*current/1000

        max_cs_power = math.sqrt(cs.phases) * \
            cs.voltage*cs.max_charge_current/1000
        min_cs_power = math.sqrt(cs.phases) * \
            cs.voltage*cs.min_charge_current/1000

        if cs_power_potential > max_cs_power:
            power_potential += max_cs_power
        elif cs_power_potential < min_cs_power:
            power_potential += 0
        else:
            power_potential += cs_power_potential

    return power_potential



# This script loads the residential_loads.csv file and plots the data.


import pandas as pd

# Print the first few rows of the dataframe
import matplotlib.pyplot as plt
# Plot the data using pandas
# data.plot()

# Load the data
data = pd.read_csv('residential_loads.csv',header=None)


desired_timescale = 15 ##### from env
simulation_length = 96 ##### from env
simulation_date = '2023-2-1 08:00:00' ##### from env
number_of_transformers = 3 ##### from env

dataset_timescale = 15
dataset_starting_date = '2022-01-01 00:00:00'


if desired_timescale > dataset_timescale:
    data = data.groupby(data.index // (desired_timescale/dataset_timescale)).max()
elif desired_timescale < dataset_timescale:
    # extend the dataset to data.shape[0] * (dataset_timescale/desired_timescale)
    # by repeating the data every (dataset_timescale/desired_timescale) rows
    data = data.loc[data.index.repeat(dataset_timescale/desired_timescale)].reset_index(drop=True)
    # data = data/ (dataset_timescale/desired_timescale)

# duplicate the data to have two years of data
data = pd.concat([data, data], ignore_index=True)

#add a date column to the dataframe
data['date'] = pd.date_range(start=dataset_starting_date, periods=data.shape[0], freq=f'{desired_timescale}T')
    
#find year of the data
year = int(dataset_starting_date.split('-')[0])
# replace the year of the simulation date with the year of the data
simulation_date = f'{year}-{simulation_date.split("-")[1]}-{simulation_date.split("-")[2]}'
print(f'simulation date: {simulation_date}')
simulation_index = data[data['date'] == simulation_date].index[0]

#select the data for the simulation date
data = data[simulation_index:simulation_index+simulation_length]

print(f'simulation index: {simulation_index}')
print(data.head())
#drop the date column
data = data.drop(columns=['date'])

print(data.columns)
print(data.shape)

for i in range(number_of_transformers):
    data['tr_'+str(i)] = data.sample(10, axis=1).sum(axis=1)

#plot data[tr_ data
data['tr_0'].plot()
data['tr_1'].plot()
data['tr_2'].plot()

#show the time of the simulation on the x-axis

plt.xlabel('Time')
plt.ylabel('Power (kW)')

# Show the plot
plt.show()

exit()


#select 7 random columns and sum them
data['sum'] = data.sample(10, axis=1).sum(axis=1)
# Plot the data using pandas
# data['sum'].plot()

#plot the daily sum of the data (aggregate every 96 rows)



data['sum'][:96].plot()


print(data.columns)
print(data.shape)

# Show the plot
plt.show()


'''
===================================
Author: Stavros Orfanoudakis 2023
===================================
'''

import numpy as np
import gurobipy as gp
from gurobipy import GRB
from gurobipy import *
import pickle


class PowerTrackingErrorrMin():
    '''
    This file contains the PowerTrackingErrorrMin class, which is used to solve the ev_city V2G problem optimally.
    '''
    algo_name = 'Optimal (Offline)'        
    def __init__(self, replay_path=None, **kwargs):        
        
        replay = pickle.load(open(replay_path, 'rb'))

        self.sim_length = replay.sim_length
        self.n_cs = replay.n_cs
        self.number_of_ports_per_cs = replay.max_n_ports
        self.n_transformers = replay.n_transformers
        self.timescale = replay.timescale
        dt = replay.timescale / 60  # time step
        # print(f'\nGurobi MIQP solver.')
        # print('Loading data...')

        tra_max_amps = replay.tra_max_amps
        tra_min_amps = replay.tra_min_amps
        cs_transformer = replay.cs_transformer
        port_max_charge_current = replay.port_max_charge_current
        port_min_charge_current = replay.port_min_charge_current
        port_max_discharge_current = replay.port_max_discharge_current
        port_min_discharge_current = replay.port_min_discharge_current
        voltages = replay.voltages / 1000  # phases included in voltage

        power_setpoints = replay.power_setpoints

        cs_ch_efficiency = replay.cs_ch_efficiency
        cs_dis_efficiency = replay.cs_dis_efficiency

        ev_max_energy = replay.ev_max_energy

        ev_max_ch_power = replay.ev_max_ch_power  # * self.dt
        ev_max_dis_power = replay.ev_max_dis_power  # * self.dt
        u = replay.u
        energy_at_arrival = replay.energy_at_arrival
        ev_arrival = replay.ev_arrival
        t_dep = replay.t_dep
        # create model
        # print('Creating Gurobi model...')
        self.m = gp.Model("ev_city")
        self.m.setParam('OutputFlag', 0)
        # self.m.setParam('MIPGap', 0.2)

        # energy of EVs t timeslot t
        energy = self.m.addVars(self.number_of_ports_per_cs,
                                self.n_cs,
                                self.sim_length,
                                vtype=GRB.CONTINUOUS,
                                name='energy')

        current_ev_dis = self.m.addVars(self.number_of_ports_per_cs,
                                        self.n_cs,
                                        self.sim_length,
                                        vtype=GRB.CONTINUOUS,
                                        name='current_ev_dis')
        current_ev_ch = self.m.addVars(self.number_of_ports_per_cs,
                                       self.n_cs,
                                       self.sim_length,
                                       vtype=GRB.CONTINUOUS,
                                       name='current_ev_ch')

        act_current_ev_dis = self.m.addVars(self.number_of_ports_per_cs,
                                            self.n_cs,
                                            self.sim_length,
                                            vtype=GRB.CONTINUOUS,
                                            name='act_current_ev_dis')
        act_current_ev_ch = self.m.addVars(self.number_of_ports_per_cs,
                                           self.n_cs,
                                           self.sim_length,
                                           vtype=GRB.CONTINUOUS,
                                           name='act_current_ev_ch')

        current_cs_ch = self.m.addVars(self.n_cs,
                                       self.sim_length,
                                       vtype=GRB.CONTINUOUS,
                                       name='current_cs_ch')

        current_cs_dis = self.m.addVars(self.n_cs,
                                        self.sim_length,
                                        vtype=GRB.CONTINUOUS,
                                        name='current_cs_dis')

        omega_ch = self.m.addVars(self.number_of_ports_per_cs,
                                  self.n_cs,
                                  self.sim_length,
                                  vtype=GRB.BINARY,
                                  name='omega_ch')
        omega_dis = self.m.addVars(self.number_of_ports_per_cs,
                                   self.n_cs,
                                   self.sim_length,
                                   vtype=GRB.BINARY,
                                   name='omega_dis')

        power_cs_ch = self.m.addVars(self.n_cs,
                                     self.sim_length,
                                     vtype=GRB.CONTINUOUS,
                                     name='power_cs_ch')

        power_cs_dis = self.m.addVars(self.n_cs,
                                      self.sim_length,
                                      vtype=GRB.CONTINUOUS,
                                      name='power_cs_dis')

        power_tr_ch = self.m.addVars(self.n_transformers,
                                     self.sim_length,
                                     vtype=GRB.CONTINUOUS,
                                     name='power_tr_ch')

        power_tr_dis = self.m.addVars(self.n_transformers,
                                      self.sim_length,
                                      vtype=GRB.CONTINUOUS,
                                      name='power_tr_dis')

        current_tr_ch = self.m.addVars(self.n_transformers,
                                       self.sim_length,
                                       vtype=GRB.CONTINUOUS,
                                       name='current_tr_ch')
        current_tr_dis = self.m.addVars(self.n_transformers,
                                        self.sim_length,
                                        vtype=GRB.CONTINUOUS,
                                        name='current_tr_dis')

        power_error = self.m.addVars(self.sim_length,
                                     vtype=GRB.CONTINUOUS,
                                     name='power_error')

        total_soc = self.m.addVars(self.sim_length,
                                   vtype=GRB.CONTINUOUS,
                                   name='total_soc')

        # Constrains
        # print('Creating constraints...')
        # transformer current and power variables
        for t in range(self.sim_length):
            for i in range(self.n_transformers):
                self.m.addConstr(current_tr_ch[i, t] == gp.quicksum(current_cs_ch[m, t]
                                                                    for m in range(self.n_cs)
                                                                    if cs_transformer[m] == i))
                self.m.addConstr(current_tr_dis[i, t] == gp.quicksum(current_cs_dis[m, t]
                                                                     for m in range(self.n_cs)
                                                                     if cs_transformer[m] == i))

                self.m.addConstr(power_tr_ch[i, t] == gp.quicksum(power_cs_ch[m, t]
                                                                  for m in range(self.n_cs)
                                                                  if cs_transformer[m] == i),
                                 name=f'power_tr_ch.{i}.{t}')
                self.m.addConstr(power_tr_dis[i, t] == gp.quicksum(power_cs_dis[m, t]
                                                                   for m in range(self.n_cs)
                                                                   if cs_transformer[m] == i),
                                 name=f'power_tr_dis.{i}.{t}')

            power_error[t] = (gp.quicksum(power_tr_ch[i, t] - power_tr_dis[i, t]
                              for i in range(self.n_transformers))
                              - power_setpoints[t])**2

            # power_error[t] = gp.quicksum(power_tr_ch[i, t] - power_tr_dis[i, t]
            #                   for i in range(self.n_transformers))

        # total_error = gp.quicksum(power_error[t] for t in range(self.sim_length)
        #                           if power_error[t] > 0)

        # CS power output constraint
        self.m.addConstrs(power_cs_ch[i, t] == (current_cs_ch[i, t] * voltages[i])
                          for i in range(self.n_cs)
                          for t in range(self.sim_length))
        self.m.addConstrs(power_cs_dis[i, t] == (current_cs_dis[i, t] * voltages[i])
                          for i in range(self.n_cs)
                          for t in range(self.sim_length))

        # transformer current output constraint (circuit breaker)
        self.m.addConstrs((current_tr_ch[i, t] - current_tr_dis[i, t] <= tra_max_amps[i, t]
                           for i in range(self.n_transformers)
                           for t in range(self.sim_length)), name='tr_current_limit_max')
        self.m.addConstrs((current_tr_ch[i, t] - current_tr_dis[i, t] >= tra_min_amps[i, t]
                           for i in range(self.n_transformers)
                           for t in range(self.sim_length)), name='tr_current_limit_min')

        # charging station total current output (sum of ports) constraint
        self.m.addConstrs((current_cs_ch[i, t] == act_current_ev_ch.sum('*', i, t)
                           for i in range(self.n_cs)
                           for t in range(self.sim_length)), name='cs_ch_current_output')
        self.m.addConstrs((current_cs_dis[i, t] == act_current_ev_dis.sum('*', i, t)
                           for i in range(self.n_cs)
                           for t in range(self.sim_length)), name='cs_dis_current_output')

        # charging station current output constraint
        self.m.addConstrs((-current_cs_dis[i, t] + current_cs_ch[i, t] >= port_max_discharge_current[i]
                           for i in range(self.n_cs)
                           for t in range(self.sim_length)), name='cs_current_dis_limit_max')
        self.m.addConstrs((-current_cs_dis[i, t] + current_cs_ch[i, t] <= port_max_charge_current[i]
                           for i in range(self.n_cs)
                           for t in range(self.sim_length)), name='cs_curent_ch_limit_max')

        self.m.addConstrs((act_current_ev_ch[p, i, t] == current_ev_ch[p, i, t] * omega_ch[p, i, t]
                           for p in range(self.number_of_ports_per_cs)
                           for i in range(self.n_cs)
                           for t in range(self.sim_length)
                           #    if u[p, i, t] == 1 and ev_arrival[p, i, t] == 0
                           ), name='act_ev_current_ch')

        self.m.addConstrs((act_current_ev_dis[p, i, t] == current_ev_dis[p, i, t] * omega_dis[p, i, t]
                           for p in range(self.number_of_ports_per_cs)
                           for i in range(self.n_cs)
                           for t in range(self.sim_length)
                           #    if u[p, i, t] == 1 and ev_arrival[p, i, t] == 0
                           ), name='act_ev_current_dis')

        # ev current output constraint
        self.m.addConstrs((current_ev_ch[p, i, t] >= port_min_charge_current[i]  # * omega_ch[p, i, t]
                           for p in range(self.number_of_ports_per_cs)
                           for i in range(self.n_cs)
                           for t in range(self.sim_length)
                           #    if u[p, i, t] == 1 and ev_arrival[p, i, t] == 0
                           ), name='ev_current_ch_limit_min')
        self.m.addConstrs((current_ev_dis[p, i, t] >= -port_min_discharge_current[i]  # * omega_dis[p, i, t]
                           for p in range(self.number_of_ports_per_cs)
                           for i in range(self.n_cs)
                           for t in range(self.sim_length)
                           #    if u[p, i, t] == 1 and ev_arrival[p, i, t] == 0
                           ), name='ev_current_dis_limit_min')

        # ev max charging current constraint
        self.m.addConstrs((current_ev_ch[p, i, t] <= min(ev_max_ch_power[p, i, t]/voltages[i], port_max_charge_current[i])
                           for p in range(self.number_of_ports_per_cs)
                           for i in range(self.n_cs)
                           for t in range(self.sim_length)
                           if u[p, i, t] == 1 and ev_arrival[p, i, t] == 0
                           ),
                          name='ev_current_ch_limit_max')

        # ev max discharging current constraint
        self.m.addConstrs((current_ev_dis[p, i, t] <= min(-ev_max_dis_power[p, i, t]/voltages[i], -port_max_discharge_current[i])
                           for p in range(self.number_of_ports_per_cs)
                           for i in range(self.n_cs)
                           for t in range(self.sim_length)
                           if u[p, i, t] == 1 and ev_arrival[p, i, t] == 0
                           ),
                          name='ev_current_dis_limit_max')

        # ev charge power if empty port constraint
        for t in range(self.sim_length):
            for i in range(self.n_cs):
                for p in range(self.number_of_ports_per_cs):
                    if u[p, i, t] == 0 or ev_arrival[p, i, t] == 1:
                        #     self.m.addLConstr((act_current_ev_ch[p, i, t] == 0),
                        #                       name=f'ev_empty_port_ch.{p}.{i}.{t}')
                        #     self.m.addLConstr((act_current_ev_dis[p, i, t] == 0),
                        #                       name=f'ev_empty_port_dis.{p}.{i}.{t}')

                        self.m.addLConstr((omega_ch[p, i, t] == 0),
                                          name=f'omega_empty_port_ch.{p}.{i}.{t}')
                        self.m.addLConstr((omega_dis[p, i, t] == 0),
                                          name=f'omega_empty_port_dis.{p}.{i}.{t}')

                    if u[p, i, t] == 0 and t_dep[p, i, t] == 0:
                        self.m.addLConstr(energy[p, i, t] == 0,
                                          name=f'ev_empty_port_energy.{p}.{i}.{t}')

        # energy of EVs after charge/discharge constraint
        for t in range(1, self.sim_length):
            for i in range(self.n_cs):
                for p in range(self.number_of_ports_per_cs):
                    if ev_arrival[p, i, t] == 1:
                        self.m.addLConstr(
                            energy[p, i, t] == energy_at_arrival[p, i, t],
                            name=f'ev_arrival_energy.{p}.{i}.{t}')

                    if u[p, i, t-1] == 1:
                        self.m.addConstr(energy[p, i, t] == (energy[p, i, t-1] +
                                                             act_current_ev_ch[p, i, t] * voltages[i] * cs_ch_efficiency[i, t] * dt -
                                                             act_current_ev_dis[p, i, t] * voltages[i] * cs_dis_efficiency[i, t] * dt),
                                         name=f'ev_energy.{p}.{i}.{t}')

        # energy level of EVs constraint
        self.m.addConstrs((energy[p, i, t] >= 0
                           for p in range(self.number_of_ports_per_cs)
                           for i in range(self.n_cs)
                           for t in range(self.sim_length)), name='ev_energy_level_min')
        self.m.addConstrs((energy[p, i, t] <= ev_max_energy[p, i, t]
                           for p in range(self.number_of_ports_per_cs)
                           for i in range(self.n_cs)
                           for t in range(self.sim_length)
                           if t_dep[p, i, t] != 1
                           ), name='ev_energy_level_max')

        # Power output of EVs constraint
        self.m.addConstrs((omega_dis[p, i, t] * omega_ch[p, i, t] == 0
                           for p in range(self.number_of_ports_per_cs)
                           for i in range(self.n_cs)
                           for t in range(self.sim_length)), name='ev_power_mode_2')

        # self.m.addConstrs((omega_dis[p, i, t] + omega_ch[p, i, t] <= 0.99
        #                    for p in range(self.number_of_ports_per_cs)
        #                    for i in range(self.n_cs)
        #                    for t in range(self.sim_length)), name='ev_power_mode_2')

        # time of departure of EVs
        for t in range(self.sim_length):
            # flag = False
            # for i in range(self.n_cs):

            # total_soc[t] = gp.quicksum(energy[p, i, t]
            #                            for p in range(self.number_of_ports_per_cs)
            #                            if t_dep[p, i, t] == 1)
            # for p in range(self.number_of_ports_per_cs):
            #     # if t_dep[p, i, t] == 1:
            #     #     # total_soc[t] += energy[p, i, t-1]
            #     #     flag = True
            # input(f'Energy at departure: {t_dep[p,i,t]}')
            # self.m.addLConstr((energy[p, i, t] >= 15),
            #                 #    ev_des_energy[p, i, t]),
            #                   name=f'ev_departure_energy.{p}.{i}.{t}')
            # else:
            #     total_soc[t] = 0
            # if not flag:
            # total_soc[t] = 0

            total_soc[t] = gp.quicksum(energy[p, i, t]
                                       for p in range(self.number_of_ports_per_cs)
                                       for i in range(self.n_cs)
                                       if t_dep[p, i, t] == 1)

        # Objective function
        # self.m.setObjective( 10000000 * power_error.sum() - total_soc.sum(),
        #                     GRB.MINIMIZE)
        self.m.setObjective(power_error.sum(),
                            GRB.MINIMIZE)

        # print constraints
        # self.m.write("model.lp")
        print(f'Optimizing....')
        self.m.params.NonConvex = 2
        self.m.optimize()

        self.act_current_ev_ch = act_current_ev_ch
        self.act_current_ev_dis = act_current_ev_dis
        self.port_max_charge_current = port_max_charge_current
        self.port_max_discharge_current = port_max_discharge_current

        # total_soc_sum = 0
        # total_error_sum = 0
        # for t in range(self.sim_length):
        #     total_soc_sum += total_soc[t].x
        #     total_error_sum += power_error[t].x
        # print(f'Total SOC: {total_soc_sum:.2f} kWh, Total error: {total_error_sum:.2f} kW')
        # for t in range(self.sim_length):
        #     print(
        #         f'------------------ Time {t} ------------------------------------------- ')
        #     print(f'Power setpoint: {power_setpoints[t]:.2f} kW')
        #     for tr in range(self.n_transformers):
        #         print(
        #             f'Power: {(power_tr_ch[tr, t].x - power_tr_dis[tr, t].x):.2f} kW')
        #         print(f'Current: {(current_tr_ch[tr, t].x-current_tr_dis[tr, t].x):.2f} A')

        #     for i in range(self.n_cs):
        #         print(
        #             f'CS.{i}: {current_cs_ch[i, t].x:.2f} {current_cs_dis[i, t].x:.2f} voltage: {voltages[i]:.2f} kV')

        #         for p in range(self.number_of_ports_per_cs):

        #             print(f'Port {p} :'
        #                   f' Energy {energy[p, i, t].x:.2f} |' +
        #                   f' Ch {current_ev_ch[p, i, t].x:.2f}' +
        #                   f' o_ch {omega_ch[p, i, t].x:2.0f}  -  ' +
        #                   f' Dis {current_ev_dis[p, i, t].x:.2f}' +
        #                   f' o_dis {omega_dis[p, i, t].x:2.0f}|' +
        #                   f' u {u[p, i, t]:4.1f}')
        # print(t_dep)
        # print(
        #     f'Is MIP?: {self.m.IsMIP}, IsMultiObj?: {self.m.IsMultiObj}, Is QCP?: {self.m.IsQCP}, Is QP?: {self.m.IsQP}')
        if self.m.status != GRB.Status.OPTIMAL:
            print(f'Optimization ended with status {self.m.status}')
            exit()

        self.get_actions()

    def get_actions(self):
        '''
        This function returns the actions of the EVs in the simulation normalized to [-1, 1]
        '''

        self.actions = np.zeros([self.number_of_ports_per_cs,
                                 self.n_cs, self.sim_length])

        for t in range(self.sim_length):
            for i in range(self.n_cs):
                for p in range(self.number_of_ports_per_cs):
                    if self.act_current_ev_ch[p, i, t].x > 0:
                        self.actions[p, i, t] = self.act_current_ev_ch[p, i, t].x  \
                            / self.port_max_charge_current[i]
                    elif self.act_current_ev_dis[p, i, t].x > 0:
                        self.actions[p, i, t] = self.act_current_ev_dis[p, i, t].x \
                            / self.port_max_discharge_current[i]

        return self.actions

    def get_action(self, env, **kwargs):
        '''
        This function returns the action for the current step of the environment.
        '''

        step = env.current_step

        return self.actions[:, :, step].T.reshape(-1)



'''
This file contains the loaders for the EV City environment.
'''

import numpy as np
import pandas as pd
import math
import datetime
import pkg_resources
import json
from typing import List, Tuple

from ev2gym.models.ev_charger import EV_Charger
from ev2gym.models.ev import EV
from ev2gym.models.transformer import Transformer

from ev2gym.utilities.utils import EV_spawner, generate_power_setpoints


def load_ev_spawn_scenarios(env) -> None:
    '''Loads the EV spawn scenarios of the simulation'''

    df_arrival_week_file = pkg_resources.resource_filename(
        'ev2gym', 'data/distribution-of-arrival.csv')
    df_arrival_weekend_file = pkg_resources.resource_filename(
        'ev2gym', 'data/distribution-of-arrival-weekend.csv')
    df_connection_time_file = pkg_resources.resource_filename(
        'ev2gym', 'data/distribution-of-connection-time.csv')
    df_energy_demand_file = pkg_resources.resource_filename(
        'ev2gym', 'data/distribution-of-energy-demand.csv')
    time_of_connection_vs_hour_file = pkg_resources.resource_filename(
        'ev2gym', 'data/time_of_connection_vs_hour.npy')

    df_req_energy_file = pkg_resources.resource_filename(
        'ev2gym', 'data/mean-demand-per-arrival.csv')
    df_time_of_stay_vs_arrival_file = pkg_resources.resource_filename(
        'ev2gym', 'data/mean-session-length-per.csv')

    ev_specs_file = pkg_resources.resource_filename(
        'ev2gym', 'data/ev_specs.json')

    env.df_arrival_week = pd.read_csv(df_arrival_week_file)  # weekdays
    env.df_arrival_weekend = pd.read_csv(df_arrival_weekend_file)  # weekends
    env.df_connection_time = pd.read_csv(
        df_connection_time_file)  # connection time
    env.df_energy_demand = pd.read_csv(df_energy_demand_file)  # energy demand
    env.time_of_connection_vs_hour = np.load(
        time_of_connection_vs_hour_file)  # time of connection vs hour

    env.df_req_energy = pd.read_csv(
        df_req_energy_file)  # energy demand per arrival
    # replace column work with workplace
    env.df_req_energy = env.df_req_energy.rename(columns={'work': 'workplace',
                                                          'home': 'private'})
    env.df_req_energy = env.df_req_energy.fillna(0)

    env.df_time_of_stay_vs_arrival = pd.read_csv(
        df_time_of_stay_vs_arrival_file)  # time of stay vs arrival
    env.df_time_of_stay_vs_arrival = env.df_time_of_stay_vs_arrival.fillna(0)
    env.df_time_of_stay_vs_arrival = env.df_time_of_stay_vs_arrival.rename(columns={'work': 'workplace',
                                                                                    'home': 'private'})

    # Load the EV specs
    if env.config['heterogeneous_ev_specs']:
        with open(ev_specs_file) as f:
            env.ev_specs = json.load(f)

        registrations = np.zeros(len(env.ev_specs.keys()))
        for i, ev_name in enumerate(env.ev_specs.keys()):
            # sum the total number of registrations
            registrations[i] = env.ev_specs[ev_name]['number_of_registrations_2023_nl']

        env.normalized_ev_registrations = registrations/registrations.sum()


def load_power_setpoints(env) -> np.ndarray:
    '''
    Loads the power setpoints of the simulation based on the day-ahead prices
    '''

    if env.load_from_replay_path:
        return env.replay.power_setpoints
    else:
        return generate_power_setpoints(env)


def generate_residential_inflexible_loads(env) -> np.ndarray:
    '''
    This function loads the inflexible loads of each transformer
    in the simulation.
    '''

    # Load the data
    data_path = pkg_resources.resource_filename(
        'ev2gym', 'data/residential_loads.csv')
    data = pd.read_csv(data_path, header=None)

    desired_timescale = env.timescale
    simulation_length = env.simulation_length
    simulation_date = env.sim_starting_date.strftime('%Y-%m-%d %H:%M:%S')
    number_of_transformers = env.number_of_transformers

    dataset_timescale = 15
    dataset_starting_date = '2022-01-01 00:00:00'

    if desired_timescale > dataset_timescale:
        data = data.groupby(
            data.index // (desired_timescale/dataset_timescale)).max()
    elif desired_timescale < dataset_timescale:
        # extend the dataset to data.shape[0] * (dataset_timescale/desired_timescale)
        # by repeating the data every (dataset_timescale/desired_timescale) rows
        data = data.loc[data.index.repeat(
            dataset_timescale/desired_timescale)].reset_index(drop=True)

    # duplicate the data to have two years of data
    data = pd.concat([data, data], ignore_index=True)

    # add a date column to the dataframe
    data['date'] = pd.date_range(
        start=dataset_starting_date, periods=data.shape[0], freq=f'{desired_timescale}min')

    # find year of the data
    year = int(dataset_starting_date.split('-')[0])
    # replace the year of the simulation date with the year of the data
    simulation_date = f'{year}-{simulation_date.split("-")[1]}-{simulation_date.split("-")[2]}'

    simulation_index = data[data['date'] == simulation_date].index[0]

    # select the data for the simulation date
    data = data[simulation_index:simulation_index+simulation_length]

    # drop the date column
    data = data.drop(columns=['date'])
    new_data = pd.DataFrame()

    for i in range(number_of_transformers):
        new_data['tr_'+str(i)] = data.sample(10, axis=1,
                                             random_state=env.tr_seed).sum(axis=1)

    # return the "tr_" columns
    return new_data.to_numpy().T


def generate_pv_generation(env) -> np.ndarray:
    '''
    This function loads the PV generation of each transformer by loading the data from a file
    and then adding minor variations to the data
    '''

    # Load the data
    data_path = pkg_resources.resource_filename(
        'ev2gym', 'data/pv_netherlands.csv')
    data = pd.read_csv(data_path, sep=',', header=0)
    data.drop(['time', 'local_time'], inplace=True, axis=1)

    desired_timescale = env.timescale
    simulation_length = env.simulation_length
    simulation_date = env.sim_starting_date.strftime('%Y-%m-%d %H:%M:%S')
    number_of_transformers = env.number_of_transformers

    dataset_timescale = 60
    dataset_starting_date = '2019-01-01 00:00:00'

    if desired_timescale > dataset_timescale:
        data = data.groupby(
            data.index // (desired_timescale/dataset_timescale)).max()
    elif desired_timescale < dataset_timescale:
        # extend the dataset to data.shape[0] * (dataset_timescale/desired_timescale)
        # by repeating the data every (dataset_timescale/desired_timescale) rows
        data = data.loc[data.index.repeat(
            dataset_timescale/desired_timescale)].reset_index(drop=True)
        # data = data/ (dataset_timescale/desired_timescale)

    # smooth data by taking the mean of every 5 rows
    data['electricity'] = data['electricity'].rolling(
        window=60//desired_timescale, min_periods=1).mean()
    # use other type of smoothing
    data['electricity'] = data['electricity'].ewm(
        span=60//desired_timescale, adjust=True).mean()

    # duplicate the data to have two years of data
    data = pd.concat([data, data], ignore_index=True)

    # add a date column to the dataframe
    data['date'] = pd.date_range(
        start=dataset_starting_date, periods=data.shape[0], freq=f'{desired_timescale}min')

    # find year of the data
    year = int(dataset_starting_date.split('-')[0])
    # replace the year of the simulation date with the year of the data
    simulation_date = f'{year}-{simulation_date.split("-")[1]}-{simulation_date.split("-")[2]}'

    simulation_index = data[data['date'] == simulation_date].index[0]

    # select the data for the simulation date
    data = data[simulation_index:simulation_index+simulation_length]

    # drop the date column
    data = data.drop(columns=['date'])
    new_data = pd.DataFrame()

    for i in range(number_of_transformers):
        new_data['tr_'+str(i)] = data * env.tr_rng.uniform(0.9, 1.1)

    return new_data.to_numpy().T


def load_transformers(env) -> List[Transformer]:
    '''Loads the transformers of the simulation
    If load_from_replay_path is None, then the transformers are created randomly

    Returns:
        - transformers: a list of transformer objects
    '''

    if env.load_from_replay_path is not None:
        return env.replay.transformers

    transformers = []

    if env.config['inflexible_loads']['include']:

        if env.scenario == 'private':
            inflexible_loads = generate_residential_inflexible_loads(env)

        # TODO add inflexible loads for public and workplace scenarios
        else:
            inflexible_loads = generate_residential_inflexible_loads(env)

    else:
        inflexible_loads = np.zeros((env.number_of_transformers,
                                    env.simulation_length))

    if env.config['solar_power']['include']:
        solar_power = generate_pv_generation(env)
    else:
        solar_power = np.zeros((env.number_of_transformers,
                                env.simulation_length))

    if env.charging_network_topology:
        # parse the topology file and create the transformers
        cs_counter = 0
        for i, tr in enumerate(env.charging_network_topology):
            cs_ids = []
            for cs in env.charging_network_topology[tr]['charging_stations']:
                cs_ids.append(cs_counter)
                cs_counter += 1
            transformer = Transformer(id=i,
                                      env=env,
                                      cs_ids=cs_ids,
                                      max_power=env.charging_network_topology[tr]['max_power'],
                                      inflexible_load=inflexible_loads[i, :],
                                      solar_power=solar_power[i, :],
                                      simulation_length=env.simulation_length
                                      )

            transformers.append(transformer)

    else:
        if env.number_of_transformers > env.cs:
            raise ValueError(
                'The number of transformers cannot be greater than the number of charging stations')
        for i in range(env.number_of_transformers):
            # get indexes where the transformer is connected
            transformer = Transformer(id=i,
                                      env=env,
                                      cs_ids=np.where(
                                          np.array(env.cs_transformers) == i)[0],
                                      max_power=env.config['transformer']['max_power'],
                                      inflexible_load=inflexible_loads[i, :],
                                      solar_power=solar_power[i, :],
                                      simulation_length=env.simulation_length
                                      )

            transformers.append(transformer)
    env.n_transformers = len(transformers)
    return transformers


def load_ev_charger_profiles(env) -> List[EV_Charger]:
    '''Loads the EV charger profiles of the simulation
    If load_from_replay_path is None, then the EV charger profiles are created randomly

    Returns:
        - ev_charger_profiles: a list of ev_charger_profile objects'''

    charging_stations = []
    if env.load_from_replay_path is not None:
        return env.replay.charging_stations

    v2g_enabled = env.config['v2g_enabled']

    if env.charging_network_topology:
        # parse the topology file and create the charging stations
        cs_counter = 0
        for i, tr in enumerate(env.charging_network_topology):
            for cs in env.charging_network_topology[tr]['charging_stations']:
                ev_charger = EV_Charger(id=cs_counter,
                                        connected_bus=0,
                                        connected_transformer=i,
                                        min_charge_current=env.charging_network_topology[tr][
                                            'charging_stations'][cs]['min_charge_current'],
                                        max_charge_current=env.charging_network_topology[tr][
                                            'charging_stations'][cs]['max_charge_current'],
                                        min_discharge_current=env.charging_network_topology[tr][
                                            'charging_stations'][cs]['min_discharge_current'],
                                        max_discharge_current=env.charging_network_topology[tr][
                                            'charging_stations'][cs]['max_discharge_current'],
                                        voltage=env.charging_network_topology[tr][
                                            'charging_stations'][cs]['voltage'],
                                        n_ports=env.charging_network_topology[tr][
                                            'charging_stations'][cs]['n_ports'],
                                        charger_type=env.charging_network_topology[tr][
                                            'charging_stations'][cs]['charger_type'],
                                        phases=env.charging_network_topology[tr]['charging_stations'][cs]['phases'],
                                        timescale=env.timescale,
                                        verbose=env.verbose,)
                cs_counter += 1
                charging_stations.append(ev_charger)
        env.cs = len(charging_stations)
        return charging_stations

    else:
        if v2g_enabled:
            max_discharge_current = env.config['charging_station']['max_discharge_current']
            min_discharge_current = env.config['charging_station']['min_discharge_current']
        else:
            max_discharge_current = 0
            min_discharge_current = 0

        for i in range(env.cs):
            ev_charger = EV_Charger(id=i,
                                    connected_bus=0,  # env.cs_buses[i],
                                    connected_transformer=env.cs_transformers[i],
                                    n_ports=env.number_of_ports_per_cs,
                                    max_charge_current=env.config['charging_station']['max_charge_current'],
                                    min_charge_current=env.config['charging_station']['min_charge_current'],
                                    max_discharge_current=max_discharge_current,
                                    min_discharge_current=min_discharge_current,
                                    phases=env.config['charging_station']['phases'],
                                    voltage=env.config['charging_station']['voltage'],
                                    timescale=env.timescale,
                                    verbose=env.verbose,)

            charging_stations.append(ev_charger)
        return charging_stations


def load_ev_profiles(env) -> List[EV]:
    '''Loads the EV profiles of the simulation
    If load_from_replay_path is None, then the EV profiles are created randomly

    Returns:
        - ev_profiles: a list of ev_profile objects'''

    if env.load_from_replay_path is None:
        
        ev_profiles = EV_spawner(env)
        while len(ev_profiles) == 0:
            ev_profiles = EV_spawner(env)
            
        return ev_profiles
    else:
        return env.replay.EVs


def load_electricity_prices(env) -> Tuple[np.ndarray, np.ndarray]:
    '''Loads the electricity prices of the simulation
    If load_from_replay_path is None, then the electricity prices are created randomly

    Returns:
        - charge_prices: a matrix of size (number of charging stations, simulation length) with the charge prices
        - discharge_prices: a matrix of size (number of charging stations, simulation length) with the discharge prices'''

    if env.load_from_replay_path is not None:
        return env.replay.charge_prices, env.replay.discharge_prices

    # else load historical prices
    file_path = pkg_resources.resource_filename(
        'ev2gym', 'data/Netherlands_day-ahead-2015-2023.csv')
    data = pd.read_csv(file_path, sep=',', header=0)
    drop_columns = ['Country', 'Datetime (Local)']
    data.drop(drop_columns, inplace=True, axis=1)
    data['year'] = pd.DatetimeIndex(data['Datetime (UTC)']).year
    data['month'] = pd.DatetimeIndex(data['Datetime (UTC)']).month
    data['day'] = pd.DatetimeIndex(data['Datetime (UTC)']).day
    data['hour'] = pd.DatetimeIndex(data['Datetime (UTC)']).hour

    # assume charge and discharge prices are the same
    # assume prices are the same for all charging stations

    charge_prices = np.zeros((env.cs, env.simulation_length))
    discharge_prices = np.zeros((env.cs, env.simulation_length))
    # for every simulation step, take the price of the corresponding hour
    sim_temp_date = env.sim_starting_date
    for i in range(env.simulation_length):

        year = sim_temp_date.year
        month = sim_temp_date.month
        day = sim_temp_date.day
        hour = sim_temp_date.hour
        # find the corresponding price
        try:
            charge_prices[:, i] = -data.loc[(data['year'] == year) & (data['month'] == month) & (data['day'] == day) & (data['hour'] == hour),
                                            'Price (EUR/MWhe)'].iloc[0]/1000  # €/kWh
            discharge_prices[:, i] = data.loc[(data['year'] == year) & (data['month'] == month) & (data['day'] == day) & (data['hour'] == hour),
                                              'Price (EUR/MWhe)'].iloc[0]/1000  # €/kWh
        except:
            print(
                'Error: no price found for the given date and hour. Using 2022 prices instead.')

            year = 2022
            if day > 28:
                day -= 1
            print("Debug:", year, month, day, hour)
            charge_prices[:, i] = -data.loc[(data['year'] == year) & (data['month'] == month) & (data['day'] == day) & (data['hour'] == hour),
                                            'Price (EUR/MWhe)'].iloc[0]/1000  # €/kWh
            discharge_prices[:, i] = data.loc[(data['year'] == year) & (data['month'] == month) & (data['day'] == day) & (data['hour'] == hour),
                                              'Price (EUR/MWhe)'].iloc[0]/1000  # €/kWh

        # step to next
        sim_temp_date = sim_temp_date + \
            datetime.timedelta(minutes=env.timescale)

    discharge_prices = discharge_prices * env.config['discharge_price_factor']
    return charge_prices, discharge_prices
